/***************************************************************************/
/* 		This code is part of Nscache - viewer of Netscape(tm)	   */
/*		browsers disk cache					   */
/*		Copyright (c) 1999,2000 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		modified 2005,2006,2007 by Harald Foerster		   */
/*		(harald_foerster@users.sourceforge.net)			   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

#include "gui.h"
#include "apassign.h"
#include "file.h"
#include "gaccel.h"
#include "gprop.h"
#include "nls.h"
#include "viewer.h"

#if GTK_MAJOR_VERSION == 1 || defined(HAVE_GLOB)
#include <gdk/gdkkeysyms.h>
#endif

#ifdef HAVE_MSIE
#include "msie.h"
#endif

/*
#include <stdio.h>
#include <time.h>
#include <gtk/gtk.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "gui-gtk.h"
#include "nscache.h"
#include "stringbuf.h"
*/

enum
{
	kColumnURL,
	kColumnFile,
	kColumnType,
	kColumnSize,
	kColumnEncoding,
	kColumnCharset,
	kColumnModify,
	kColumnAccess,
	kColumnExpire,
	kColumnsTotal,
#ifdef CONFIG_GTKCLIST_GTKCTREE
	kNumberOfColumns = kColumnsTotal,
#else
	kColumnRowData = kColumnsTotal,
	kColumnPixMime,
	kColumnPixClose = kColumnPixMime,
	kColumnPixOpen,
	kNumberOfColumns,
#endif
	kColumnInfoLastIndex = kColumnsTotal - 1
};



#ifdef CONFIG_SWAP_COLUMNS

#ifdef CONFIG_GTKCLIST_GTKCTREE
#define MainSetColumnEntryArray(array, colnr, str) \
	(array)[column_tab_array[colnr].position] = (str)
#else
#define MainSetColumnEntryArray(array, colnr, str) \
	(array)[column_tab_array[colnr].initcolnr] = (str)
#endif

#else /* ! CONFIG_SWAP_COLUMNS */

#define MainSetColumnEntryArray(array, colnr, str) \
	(array)[colnr] = (str)

#endif /* #ifdef CONFIG_SWAP_COLUMNS .. #else */

#ifdef CONFIG_GTKCLIST_GTKCTREE
#define MainSetColumnTitle(nbook_page, index)			\
{								\
	GtkWidget *label =					\
	guitl_label_new_gettext_utf8_from_str(			\
				column_info[index].label);	\
	if ((nbook_page) == kTreeList)				\
	{							\
		GuiTreeSetColumnWidget(				\
			&(notebook_list[kTreeList].store.tree),	\
					(index), label);	\
	}							\
	else							\
	{							\
		GuiListSetColumnWidget(				\
			&(notebook_list[kSortedList].store.list),\
					(index), label);	\
	}							\
}
#else /* ! CONFIG_GTKCLIST_GTKCTREE */
#define MainSetColumnTitle(nbook_page, index)			\
{								\
	gchar *str =						\
	guitl_gettext_utf8_from_str(column_info[index].label);	\
	if ((nbook_page) == kTreeList)				\
	{							\
		GuiTreeSetColumnTitle(				\
			&(notebook_list[kTreeList].store.tree),	\
						(index), str);	\
	}							\
	else							\
	{							\
		GuiListSetColumnTitle(				\
			&(notebook_list[kSortedList].store.list),\
						(index), str);	\
	}							\
	utf8_free(str);						\
}
#endif /* #ifdef CONFIG_GTKCLIST_GTKCTREE .. #else */

#ifdef CONFIG_GTKCLIST_GTKCTREE
#define MainListNew(list) \
	GuiListInit((list), kNumberOfColumns)
#else
#define MainListNew(list) \
{ \
	(list)->slist = \
		gtk_list_store_new (kNumberOfColumns - 1, \
			G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, \
			G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, \
			G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, \
			G_TYPE_POINTER, GDK_TYPE_PIXBUF); \
	GuiListInit((list), kColumnRowData); \
}
#endif

#ifdef CONFIG_GTKCLIST_GTKCTREE
#define MainListAppend(list, str, icon, data) \
{ \
	gint row; \
	row = gtk_clist_append((list)->clist, (str)); \
	gtk_clist_set_pixtext((list)->clist, row, 0, (str)[0], \
			3, (icon)->pixmap, (icon)->mask); \
	gtk_clist_set_row_data((list)->clist, row, (data)); \
}
#else
#define MainListAppend(list, str, icon, data) \
{ \
	GtkTreeIter iter; \
	gtk_list_store_append((list)->slist, &iter); \
	gtk_list_store_set((list)->slist, &iter, \
		kColumnURL, (str)[0], kColumnFile, (str)[1], \
		kColumnType, (str)[2], kColumnSize, (str)[3], \
		kColumnEncoding, (str)[4], kColumnCharset, (str)[5], \
		kColumnModify, (str)[6], kColumnAccess, (str)[7], \
		kColumnExpire, (str)[8], \
		kColumnPixMime, (icon)->pixbuf, kColumnRowData, (data), -1); \
}
#endif

#ifdef CONFIG_GTKCLIST_GTKCTREE
#define MainTreeNew(tree) \
	GuiTreeInit((tree), kNumberOfColumns, kColumnURL)
#else
#define MainTreeNew(tree) \
{ \
	(tree)->stree = \
		gtk_tree_store_new (kNumberOfColumns, \
			G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, \
			G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, \
			G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, \
			G_TYPE_POINTER, GDK_TYPE_PIXBUF, GDK_TYPE_PIXBUF); \
	GuiTreeInit((tree), kColumnRowData, kColumnPixMime); \
}
#endif

#ifdef CONFIG_GTKCLIST_GTKCTREE
#define MainTreeAppendNode(tree, parent, str, icon_close, icon_open, child) \
{ \
	char *centry[9]; \
	centry[0] = (str); \
	centry[1] = centry[2] = centry[3] = centry[4] = \
	centry[5] = centry[6] = centry[7] = centry[8] = NULL; \
	(child) = gtk_ctree_insert_node((tree)->ctree, (parent), NULL, \
				centry, 8, \
				(icon_close).pixmap, (icon_close).mask, \
				(icon_open).pixmap, (icon_open).mask, \
				FALSE, ((parent) == NULL)); \
}
#else
#define MainTreeAppendNode(tree, parent, str, icon_close, icon_open, child) \
{ \
	(child) = g_malloc(sizeof(GtkTreeIter)); \
	gtk_tree_store_append((tree)->stree, (child), (parent)); \
	gtk_tree_store_set ((tree)->stree, (child), kColumnURL, (str), \
				kColumnPixClose, (icon_close).pixbuf, \
				kColumnPixOpen, (icon_open).pixbuf, -1); \
}
#endif

#ifdef CONFIG_GTKCLIST_GTKCTREE
#define MainTreeInsertNode(tree, parent, str, icon, data) \
{ \
	GtkCTreeNode *node; \
	node = gtk_ctree_insert_node((tree)->ctree, (parent), NULL, \
			(str), 8, (icon)->pixmap, (icon)->mask, \
			(icon)->pixmap, (icon)->mask, FALSE, FALSE); \
	gtk_ctree_node_set_row_data((tree)->ctree, node, (data)); \
}
#else
#define MainTreeInsertNode(tree, parent, str, icon, data) \
{ \
	GtkTreeIter iter; \
	gtk_tree_store_append((tree)->stree, &iter, (parent)); \
	gtk_tree_store_set((tree)->stree, &iter, \
		kColumnURL, (str)[0], kColumnFile, (str)[1], \
		kColumnType, (str)[2], kColumnSize, (str)[3], \
		kColumnEncoding, (str)[4], kColumnCharset, (str)[5], \
		kColumnModify, (str)[6], kColumnAccess, (str)[7], \
		kColumnExpire, (str)[8], \
		kColumnPixMime, (icon)->pixbuf, kColumnRowData, (data), -1); \
}
#endif

#ifdef CONFIG_GTKCLIST_GTKCTREE
#define MainTreeExpandTopLevelNodes(tree) {;}
#else
#define MainTreeExpandTopLevelNodes(tree) \
{ \
	GtkTreeIter iter; \
	GtkTreeModel *model = GTK_TREE_MODEL((tree)->stree); \
	gint row = -1; \
	while (gtk_tree_model_iter_nth_child(model, &iter, NULL, ++row)) \
	{ \
		GtkTreePath *path = gtk_tree_model_get_path(model, &iter); \
		gtk_tree_view_expand_row((tree)->view,  path,  FALSE); \
		gtk_tree_path_free(path); \
	} \
}
#endif

static const char invalid_filename_str[] = gettext_nop("Invalid filename!");

#if GTK_MAJOR_VERSION == 1
#define gui_fname_error(fname)	gui_strerror(fname)
#endif

#if GTK_MAJOR_VERSION == 2

/*
	----------------------------------------------------
	 Problems using GTK+-2.2.10 with WindowMaker 0.91.0
	   and XFree86 Version 3.3.6 Protocol Version 11
	----------------------------------------------------

	Sometimes no 'focus-in-event' was received. So I use
	for GTK+-2.0 the 'property-notify-', 'map-', 'unmap-'
	and 'destroy' events, too.
	This seems to work well, but requires two additional
	lists of the window order (mapped and unmapped).
	Maybe this is buggy using other configurations ...
*/

#ifdef FOCUS_IN_BUGFIX
#define _NET_WM_STATE		"_NET_WM_STATE"
#endif

#endif /* GTK_MAJOR_VERSION == 2 */


enum
{
	kMessageBufferSize	= 4096
};

enum
{
	SEL_TYPE_FIRST		= 1,
	SEL_TYPE_STRING		= SEL_TYPE_FIRST,
	SEL_TYPE_TEXT,
	SEL_TYPE_COMPOUND_TEXT,
	SEL_TYPE_LAST		= SEL_TYPE_COMPOUND_TEXT
};


static const GtkTargetEntry targetlist[] =
{
	{ "STRING", 0, SEL_TYPE_STRING },
	{ "TEXT", 0, SEL_TYPE_TEXT },
	{ "COMPOUND_TEXT", 0, SEL_TYPE_COMPOUND_TEXT }
};


typedef struct
{
	ListSelection	*toplist;	/* list in topwin */
	GtkWidget	*topwin;	/* window with focus */
	GtkWidget	*mainwin;	/* this (gui.c) window */
#ifdef FOCUS_IN_BUGFIX
	GdkAtom		_net_wm_state;	/* seems to be WM  top window property */
	GSList		*ismapped;	/* order of mapped windows (top first) */
	GSList		*unmapped;	/* reverse order of unmapped windows */
#endif

} Windata;

static Windata	windata;

/* both kTreeList and kSortedList */
ListSelection	notebook_list[kNotebookPages];

/* either kTreeList or kSortedList */
ListSelection	*mainlist;

GSList *list_selection_chain = NULL;

static GtkEntry  *entry_db;
static GtkWidget *nscache_type;

static GtkWidget *info_dialog = NULL;
static char *selection_urlstr = NULL;
static gboolean nscache_gui_started = FALSE;


struct msg_tmr
{
	guint		tmr_id;
	guint		hdl_id;
	GtkLabel	*label;
	GtkWidget	*frame;
};

static struct msg_tmr status_msg, error_msg;


enum
{
	kNodeInfoFormatSingleLine,
	kNodeInfoFormatMultiLine
};

struct node_info
{
	unsigned int	servicenum;
	unsigned int	domainnum;
	unsigned int	foldernum;
	unsigned int	filenum;
	unsigned long	contentsize;
};

typedef struct
{
	GtkWidget	*left;
	GtkWidget	*right;

} ColumnArrows;

typedef struct
{
	gint		index;
	gint		position;
	gboolean	visible;
	GtkWidget	*menu;
#ifdef CONFIG_GTKCLIST_GTKCTREE
#ifdef CONFIG_COLUMN_BUTTONS
	ColumnArrows	arrow[kNotebookPages];
#endif
#else
#ifdef CONFIG_COLUMN_POPUP
	gint		initcolnr;
#endif
#endif

} ColumnTab;


#ifdef CONFIG_SWAP_COLUMNS

typedef struct
{
#ifdef CONFIG_COLUMN_POPUP
	GtkWidget *menu;
#else
	gboolean  is_active;
#endif
	GtkWidget *collapse;
	GtkWidget *expand;

} CollapseExpandWidgets;

#ifdef CONFIG_COLUMN_BUTTONS
static CollapseExpandWidgets button_collexp;
#endif

#ifdef CONFIG_COLUMN_POPUP

static CollapseExpandWidgets popup_collexp;

static struct popup_colswap
{
	GtkWidget *menu;
	GtkWidget *left;
	GtkWidget *right;
	ColumnTab *selected;

} popup_colswap;

#endif /* CONFIG_COLUMN_POPUP */

#endif /* CONFIG_SWAP_COLUMNS */


typedef struct
{
	gchar		*label;
	gint		 show_column_prop;
	gint		 tree_width_prop;
	gint		 list_width_prop;
	gint		 colnrprop;
	gint		 accel_id;
	gint  		 dwidth;
	GtkJustification justify;

} ColumnInfo;

static ColumnTab	column_tab_array[kColumnsTotal];
static ColumnTab	*column_tab[kColumnsTotal];

static const ColumnInfo column_info[kColumnsTotal] =
{
	{gettext_nop("Cache URL entry"),
		kPropBoolShowColUrl, kPropIntTreeColWidthUrl,
		kPropIntListColWidthUrl, kPropIntColNumUrl,
		kGAccelColumnURL, 200, GTK_JUSTIFY_LEFT},
	{gettext_nop("Local file"),
		kPropBoolShowColFile, kPropIntTreeColWidthFile,
		kPropIntListColWidthFile, kPropIntColNumFile,
		kGAccelColumnFile, 150, GTK_JUSTIFY_LEFT}, 
	{gettext_nop("Type"),
		kPropBoolShowColType, kPropIntTreeColWidthType,
		kPropIntListColWidthType, kPropIntColNumType,
		kGAccelColumnType, 90, GTK_JUSTIFY_LEFT},
	{gettext_nop("Size"),
		kPropBoolShowColSize, kPropIntTreeColWidthSize,
		kPropIntListColWidthSize, kPropIntColNumSize,
		kGAccelColumnSize, 70, GTK_JUSTIFY_RIGHT},
	{gettext_nop("Encoding"),
		kPropBoolShowColEnc, kPropIntTreeColWidthEnc,
		kPropIntListColWidthEnc, kPropIntColNumEnc,
		kGAccelColumnEncoding, 90, GTK_JUSTIFY_LEFT},
	{gettext_nop("Charset"),
		kPropBoolShowColCharset, kPropIntTreeColWidthCharset,
		kPropIntListColWidthCharset, kPropIntColNumCharset,
		kGAccelColumnCharset, 90, GTK_JUSTIFY_LEFT},
	{gettext_nop("Mod. time"),
		kPropBoolShowColMdtm, kPropIntTreeColWidthMdtm,
		kPropIntListColWidthMdtm, kPropIntColNumMdtm,
		kGAccelColumnMod, 90, GTK_JUSTIFY_LEFT},
	{gettext_nop("Access time"),
		kPropBoolShowColAtm, kPropIntTreeColWidthAtm,
		kPropIntListColWidthAtm, kPropIntColNumAtm,
		kGAccelColumnAcc, 90, GTK_JUSTIFY_LEFT},
	{gettext_nop("Expires"),
		kPropBoolShowColExp, kPropIntTreeColWidthExp,
		kPropIntListColWidthExp, kPropIntColNumExp,
		kGAccelColumnExp, 90, GTK_JUSTIFY_LEFT}

}; /* static const ColumnInfo column_info[kColumnsTotal] */

/*
	-----------------------
	 CHECK BUTTONS IN MENU
	-----------------------
*/

enum
{
	kOptionConvertToLower,
#ifndef CACHE_READONLY
	kOptionReadOnly,
#endif
	kOptionSaveConfigOnExit,
	kOptionEntriesTotal
};

static struct gui_option
{
	gint		active;
	gint		prop;
	GtkWidget	*widget;

} gui_option[] = {

	{ FALSE, kPropBoolOptFnameToLower },
#ifndef CACHE_READONLY
	{ FALSE, kPropBoolOptReadOnly},
#endif
	{  TRUE, kPropBoolOptSaveConfig }
};

/*
	------------------------------------------------
	 enable/disable CheckButton in SaveFolderDialog
	------------------------------------------------
*/

#ifdef HAVE_LINK
static gint hardlink_enable;
#endif


/*
	------------
	 PROTOTYPES
	------------
*/

/* File Menu */
static void Browse(GtkWidget *w, gpointer data);
static void PropertiesDialog(GtkWidget *w, gpointer data);
static void SaveRC(GtkWidget *w, gpointer data);
static void Quit(GtkWidget *w, gpointer data);

/* View Menu */
static void ViewInfo(GtkWidget *w, ListSelection **selected);
static void GenViewerMenu(GtkWidget *submenu, int viewer_type);

/* Edit Menu */
static void SaveAs(GtkWidget *w, ListSelection **selected);

#ifndef CACHE_READONLY
static void DeleteDialog(GtkWidget *w, ListSelection **selected);
#endif

static void SelectionSetURL(GtkWidget *w, ListSelection **selected);
static void NodeInfo(GtkWidget *w, ListSelection **selected);
static void RdOnly(GtkWidget *w, ListSelection **selected);
static void Writable(GtkWidget *w, ListSelection **selected);

/* Options Menu */
static void ToggleBool(GtkWidget *w, struct gui_option *opt);

/* Help Menu */
static void About(GtkWidget *w, gpointer data);

/* Misc Prototypes */
static void Open(GtkEntry *entry);
static void update_entry_info(ListSelection *list);
static void gui_msg(const gchar *str);
static void gui_window_resize(GtkWidget *w, GtkAllocation *sallocation, gpointer data);

#ifdef CONFIG_SWAP_COLUMNS
static void
gui_test_expand_recursive(GuiTreeStore *tree,nscache_record_t *rec, gboolean *can_expand);
static void
gui_collapse_expand_set_sensitive(ListSelection *list, CollapseExpandWidgets *cew);
#endif

/* called from gaccel.c and gui.c */
void AdjustMenu(GtkWidget *w, ListSelection **selected);


/*
	--------------
	 MENU WIDGETS
	--------------
*/

enum
{
	kMenuViewFile,
	kMenuViewURL,
	kMenuFind,

	kMenuSave,
#ifndef CACHE_READONLY
	kMenuDelete,
#endif
	kMenuClipboard,
	kMenuReadOnly,
	kMenuReadWrite,
	kMenuItemsTotal
};

static GtkWidget *gui_menu_widget[kMenuItemsTotal];
static GtkWidget *gui_popup_widget[kMenuItemsTotal];

GtkWidget *gui_popup_menu;

static const char mn_save_file_str[]	= gettext_nop("Save to file ...");
static const char mn_delete_str[]	= gettext_nop("Delete file ...");
static const char mn_clipboard_str[]	= gettext_nop("Copy url to clipboard");
static const char mn_entry_info_str[]	= gettext_nop("Entry informations ...");
static const char mn_read_only_str[]	= gettext_nop("read-only permissions");
static const char mn_read_write_str[]	= gettext_nop("read/write permissions");
static const char mn_view_file_str[]	= gettext_nop("View file with");
static const char mn_view_url_str[]	= gettext_nop("View URL with");
static const char mn_find_str[]		= gettext_nop("Find ...");


/*
	------------
	 MENU TYPES
	------------
*/

enum
{
	kMTypeEndOfList,
	kMTypeItem,
	kMTypeCheckItem,
	kMTypeSeparator,
	kMTypeSubmenu,
	kMTypeSubmenuCheckItem

};

/*
	------------
	 MENU ITEMS
	------------
*/

typedef struct
{
	gint		type;		/* see above */
	const gchar	*label;		/* name to display */
	gint		accel_id;	/* accel key */
	GtkWidget	**storage;	/* gui_menu_widget */
	GtkSignalFunc	callback;	/* function pointer */
	gpointer	data;		/* data for it */

} GuiMenuItem;

static const GuiMenuItem menuFile[] =
{
	{kMTypeItem, gettext_nop("Open cache index file ..."),
			kGAccelFileOpen, NULL, (GtkSignalFunc) Browse, NULL},
	{kMTypeItem, gettext_nop("Properties ..."),
			kGAccelFileProp, NULL, (GtkSignalFunc) PropertiesDialog, NULL},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Save rc file"),
			kGAccelFileSaveRC, NULL, (GtkSignalFunc) SaveRC, NULL},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Quit"),
			kGAccelFileQuit, NULL, (GtkSignalFunc) Quit, NULL},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const GuiMenuItem menuFile[] */

static const GuiMenuItem menuView[] =
{
	{kMTypeItem, mn_entry_info_str,
			kGAccelViewEntryInfo, NULL,
			(GtkSignalFunc) ViewInfo, &windata.toplist},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeSubmenuCheckItem, gettext_nop("Columns"),
			kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeSubmenu, mn_view_file_str,
			kGAccelKeyNone, &gui_menu_widget[kMenuViewFile],
			(GtkSignalFunc) GenViewerMenu, (gpointer) VIEWER_FILE},
	{kMTypeSubmenu, mn_view_url_str,
			kGAccelKeyNone, &gui_menu_widget[kMenuViewURL],
			(GtkSignalFunc) GenViewerMenu, (gpointer) VIEWER_URL},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_find_str,
			kGAccelViewFind, &gui_menu_widget[kMenuFind],
			(GtkSignalFunc) search_dlg_open, NULL},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const GuiMenuItem menuView[] = */

static const GuiMenuItem menuEdit[] =
{
	{kMTypeItem, mn_save_file_str,
			kGAccelEditSave, &gui_menu_widget[kMenuSave],
			(GtkSignalFunc) SaveAs, &windata.toplist},
#ifndef CACHE_READONLY
	{kMTypeItem, mn_delete_str,
			kGAccelEditDelete, &gui_menu_widget[kMenuDelete],
			(GtkSignalFunc) DeleteDialog, &windata.toplist},
#endif
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_clipboard_str,
			kGAccelEditClipboard, &gui_menu_widget[kMenuClipboard],
			(GtkSignalFunc) SelectionSetURL, &windata.toplist},
	{kMTypeItem, gettext_nop("Show entry informations"),
			kGAccelEditShowInfo, NULL,
			(GtkSignalFunc) NodeInfo, &windata.toplist},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_read_only_str,
			kGAccelEditReadOnly, &gui_menu_widget[kMenuReadOnly],
			(GtkSignalFunc) RdOnly, &windata.toplist},
	{kMTypeItem, mn_read_write_str,
			kGAccelEditReadWrite, &gui_menu_widget[kMenuReadWrite],
			(GtkSignalFunc) Writable, &windata.toplist},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const GuiMenuItem menuEdit[] */

static const GuiMenuItem menuOptions[] =
{
	{kMTypeItem, gettext_nop("Viewers setup ..."),
			kGAccelOptViewers, NULL, (GtkSignalFunc) apassign_run, NULL},
	{kMTypeItem, gettext_nop("Keyboard ..."),
			kGAccelOptAccel, NULL, (GtkSignalFunc) gaccel_assign_dialog, NULL},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeCheckItem, gettext_nop("Convert filenames to lowercase"),
			kGAccelOptCase, &gui_option[kOptionConvertToLower].widget,
			(GtkSignalFunc) ToggleBool, &gui_option[kOptionConvertToLower]},

#ifndef CACHE_READONLY
	{kMTypeCheckItem, gettext_nop("Readonly"),
			kGAccelOptRdOnly, &gui_option[kOptionReadOnly].widget,
			(GtkSignalFunc) ToggleBool, &gui_option[kOptionReadOnly]},
#endif
	{kMTypeCheckItem, gettext_nop("Save rc on exit"),
			kGAccelOptSave, &gui_option[kOptionSaveConfigOnExit].widget,
			(GtkSignalFunc) ToggleBool, &gui_option[kOptionSaveConfigOnExit]},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const GuiMenuItem menuOptions[] */

static const GuiMenuItem menuHelp[] =
{
	{kMTypeItem, gettext_nop("About ..."),
		kGAccelHelpAbout, NULL, (GtkSignalFunc) About, NULL},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const GuiMenuItem menuHelp[] */

static const GuiMenuItem menuPopup[] =
{
	{kMTypeItem, mn_save_file_str,
			kGAccelKeyNone, &gui_popup_widget[kMenuSave],
			(GtkSignalFunc) SaveAs, &windata.toplist},
#ifndef CACHE_READONLY
	{kMTypeItem, mn_delete_str,
			kGAccelKeyNone, &gui_popup_widget[kMenuDelete],
			(GtkSignalFunc) DeleteDialog, &windata.toplist},
#endif
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_clipboard_str,
			kGAccelKeyNone, &gui_popup_widget[kMenuClipboard],
			(GtkSignalFunc) SelectionSetURL, &windata.toplist},
	{kMTypeItem, mn_entry_info_str,
			kGAccelKeyNone, NULL,
			(GtkSignalFunc) ViewInfo, &windata.toplist},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_read_only_str,
			kGAccelKeyNone, &gui_popup_widget[kMenuReadOnly],
			(GtkSignalFunc) RdOnly, &windata.toplist},
	{kMTypeItem, mn_read_write_str,
			kGAccelKeyNone, &gui_popup_widget[kMenuReadWrite],
			(GtkSignalFunc) Writable, &windata.toplist},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeSubmenu, mn_view_file_str,
			kGAccelKeyNone, &gui_popup_widget[kMenuViewFile],
			(GtkSignalFunc) GenViewerMenu, (gpointer) VIEWER_FILE},
	{kMTypeSubmenu, mn_view_url_str,
			kGAccelKeyNone, &gui_popup_widget[kMenuViewURL],
			(GtkSignalFunc) GenViewerMenu, (gpointer) VIEWER_URL},
	{kMTypeSeparator, NULL, kGAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_find_str,
			kGAccelKeyNone, &gui_popup_widget[kMenuFind],
			(GtkSignalFunc) search_dlg_open, NULL},
	{kMTypeEndOfList, NULL, kGAccelKeyNone, NULL, NULL, NULL}

}; /* static const GuiMenuItem menuPopup[] */



#ifdef FOCUS_IN_BUGFIX

static void gui_window_topped(Windata *windat)
{
	GtkWidget *window;

	window = (windat->ismapped == NULL) ?
				windat->unmapped->data :
					windat->ismapped->data;
	if (window != windat->topwin)
	{
		ListSelection *list;

		windat->topwin = window;

		list = GuiObjectGetUserData(window);

		if (list)
		{
			windat->toplist = list;

			update_entry_info(list);
		}
	}
}

static void gui_window_destroyed(GtkWidget *window, Windata *windat)
{
	if (window != windat->mainwin)
	{
		GSList **list = &windat->ismapped;

		if (g_slist_find(*list, window) == NULL)
		{
			list = &windat->unmapped;
		}

		*list = g_slist_remove(*list, window);

		gui_window_topped(windat);
	}
}

static gint gui_window_focus_property(GtkWidget *window,
					GdkEventProperty *e, Windata *windat)
{
	if (window != windat->topwin &&
		(e->type == GDK_FOCUS_CHANGE ||
			GDK_ATOM_TO_POINTER(e->atom) ==
				GDK_ATOM_TO_POINTER(windat->_net_wm_state)))
	{
		GSList *top = g_slist_find(windat->ismapped, window);

		if (top)
		{
			windat->ismapped = g_slist_remove_link(windat->ismapped, top);
			windat->ismapped = g_slist_concat(top, windat->ismapped);

			gui_window_topped(windat);
		}
	}

	return FALSE;
}

static gint gui_window_map(GtkWidget *window,
				GdkEvent *e, Windata *windat)
{
	GSList *top = g_slist_find(windat->unmapped, window);

	if (top)
	{
		windat->unmapped = g_slist_remove_link(windat->unmapped, top);
		windat->ismapped = g_slist_concat(windat->ismapped, top);

		gui_window_topped(windat);
	}

	else
	{
		windat->ismapped = g_slist_prepend(windat->ismapped, window);
	}

	return FALSE;
}

static gint gui_window_unmap(GtkWidget *window,
				GdkEvent *e, Windata *windat)
{
	GSList *top = g_slist_find(windat->ismapped, window);

	if (top)
	{
		windat->ismapped = g_slist_remove_link(windat->ismapped, top);
		windat->unmapped = g_slist_concat(top, windat->unmapped);

		gui_window_topped(windat);
	}

	return FALSE;
}

#else /* ! FOCUS_IN_BUGFIX */

#if GTK_MAJOR_VERSION > 1
static void gui_window_destroyed(GtkWidget *window, Windata *windat)
{
	if (window == windat->topwin)
	{
		windat->topwin  = NULL;
		windat->toplist = NULL;
	}
}
#endif /* GTK_MAJOR_VERSION > 1 */

static gint gui_window_focus_in(GtkWidget *window,
				GdkEventFocus *e, Windata *windat)
{
	if (window != windat->topwin)
	{
		ListSelection *list;

		windat->topwin = window;

		list = GuiObjectGetUserData(window);

		if (list)
		{
			windat->toplist = list;

			update_entry_info(list);
		}
	}

	return FALSE;
}

#endif /* #ifdef FOCUS_IN_BUGFIX ... #else */


static void gui_window_setup(GtkWidget *widget, const gchar *title, gint prop)
{
	gint		width, height;
	gchar		*str;
	GtkWindow	*window;

	width  = -1;
	height = -1;
	gprop_get_int(prop, &width);
	gprop_get_int(prop + 1,&height);

	window = GTK_WINDOW(widget);

	str = guitl_gettext_utf8_from_str(title);
	gtk_window_set_title(window, str);
	utf8_free(str);

	GuiWindowSetResizable(window, TRUE);
	gtk_window_set_default_size(window, width, height);

	GuiSignalConnect(window, "size-allocate", gui_window_resize, (gpointer) prop);

} /* static void gui_window_setup(GtkWidget *widget, const gchar *title, gint prop) */

GtkWidget *gui_window_new(const gchar *title, gint prop)
{
	GtkWidget *window;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gui_window_setup(window, title, prop);

#ifdef FOCUS_IN_BUGFIX

	gtk_widget_set_events(window, GDK_FOCUS_CHANGE_MASK |
					GDK_PROPERTY_CHANGE_MASK);

	GuiSignalConnect(window, "focus-in-event",
				gui_window_focus_property, &windata);

	GuiSignalConnect(window, "property-notify-event",
				gui_window_focus_property, &windata);

	GuiSignalConnect(window, "map-event", gui_window_map, &windata);
	GuiSignalConnect(window, "unmap-event", gui_window_unmap, &windata);
	GuiSignalConnect(window, "destroy", gui_window_destroyed, &windata);

#else /* ! FOCUS_IN_BUGFIX */

#if GTK_MAJOR_VERSION > 1
	GuiSignalConnect(window, "destroy", gui_window_destroyed, &windata);
#endif
	GuiSignalConnect(window, "focus-in-event", gui_window_focus_in, &windata);
#endif

	return window;

} /* GtkWidget *gui_window_new(const gchar *title, gint prop) */

#ifdef CONFIG_GTKCLIST_GTKCTREE
static void gui_clist_freeze_or_thaw(ListSelection *list, gpointer func)
{
	void (*freeze_or_thaw)(GtkCList*) = func;

	(*freeze_or_thaw)((list->type == kTreeList) ?
				GTK_CLIST(list->store.tree.ctree) :
						list->store.list.clist);
}
#endif /* CONFIG_GTKCLIST_GTKCTREE */

static void compute_node_info_recursive(GuiTreeStore *tree,
				nscache_record_t *rec, gpointer data)
{
	struct node_info *nfo = data;

	if (rec)
	{
		nfo->filenum++;
		nfo->contentsize += rec->content_length;
	}

	else 
	{
		gint level = GuiTreeGetLevel(tree);

		if (level == kTreeLevelService)
		{
			nfo->servicenum++;
		}

		else if (level == kTreeLevelDomain)
		{
			nfo->domainnum++;
		}

		else
		{
			nfo->foldernum++;
		}
	}
}

static void compute_node_info(ListSelection *list, struct node_info *nfo)
{
	nfo->servicenum		= 0;
	nfo->domainnum		= 0;
	nfo->foldernum		= 0;
	nfo->filenum 		= 0;
	nfo->contentsize	= 0;

	if (list && list->nscrec)
	{
		nfo->filenum = 1;
		nfo->contentsize = list->nscrec->content_length;
	}

	else
	{
		GuiTreeStore	*tree;
		gboolean	selected;

		if (list == NULL)
		{
			tree = &(notebook_list[kTreeList].store.tree);
			selected = FALSE;
		}

		else
		{
			if (list->name == NULL || list->type != kTreeList)
			{
				return;
			}

			tree = &(list->store.tree);
			selected = TRUE;
		}

		guitl_tree_recursive(tree, selected,
			(GuiTreeStoreRecursiveFunc) compute_node_info_recursive, nfo);
	}
}

static int get_node_info(char *buf, int sz, int fmt, ListSelection *list)
{
	static const char *format_str[] = 
	{
		"%s: %u  %s: %u  %s: %u  %s: %u  %s: %lu %s",
		"%s: %u\n%s: %u\n%s: %u\n%s: %u\n%s: %lu %s"
	};

	int			len;
	char			*byte_str;
	unsigned long		size;
	struct node_info	nfo;

	compute_node_info(list, &nfo);

	size = nfo.contentsize;

	if (size >= 1024)
	{
		size /= 1024;
		byte_str = gettext("kB");
	}

	else
	{
		byte_str = gettext("Bytes");
	}

	len = g_snprintf(buf, sz, format_str[fmt],
		gettext("Services"), nfo.servicenum,
			gettext("Domains"), nfo.domainnum,
				gettext("Folders"), nfo.foldernum,
					gettext("Files"), nfo.filenum,
						gettext("Size"), size, byte_str);
	buf[sz - 1] = '\0';

	return len;
}

static void print_node_info(ListSelection *list)
{
	char pom[kMessageBufferSize];

	get_node_info(pom, sizeof(pom), kNodeInfoFormatSingleLine, list);
	gui_msg(pom);
}

static gint msg_clr(struct msg_tmr *msg)
{
	if (msg->tmr_id)
	{
		if (msg->hdl_id)
		{
			GuiSignalHandlerBlock(msg->label, msg->hdl_id);
			gtk_widget_queue_draw(msg->frame);
		}

		msg->tmr_id = 0;
	}

	gtk_label_set_text(msg->label, NULL);

	return FALSE;
}

static void msg_set(struct msg_tmr *msg, const gchar *str)
{
	if (msg->tmr_id)
	{
		GuiTimeoutRemove(msg->tmr_id);

		if (str == NULL)
		{
			msg_clr(msg);
			return;
		}
	}

	else if (str == NULL)
	{
		return;
	}

	else if (msg->hdl_id)
	{
		GuiSignalHandlerUnblock(msg->label, msg->hdl_id);
	}

	msg->tmr_id = GuiTimeoutAdd(10000, (GtkFunction) msg_clr, msg);
	gtk_label_set_text(msg->label, str);
}

static void gui_msg(const gchar *str)
{
	gchar *s;

	s = utf8_from_str(str);
	msg_set(&status_msg, s);
	utf8_free(s);
}

void gui_err(const gchar *str)
{
	if (nscache_gui_started == FALSE)
	{
		if (str)
		{
			fprintf(stderr, "%s\n", str);
		}
	}

	else if (str == NULL)
	{
		msg_set(&error_msg, str);
	}

	else
	{
		gchar *s;

		s = utf8_from_str(str);
		msg_set(&error_msg, s);
		utf8_free(s);

		gdk_beep();
	}
}

void gui_errfmt(const gchar *format, ...)
{
	gchar *errmsg;
	gchar pom[kMessageBufferSize];

	errmsg = (gchar *) format;

	if (format)
	{
		va_list args;

		va_start(args, format);
		g_vsnprintf(pom, sizeof(pom), format, args);
		va_end(args);

		pom[sizeof(pom) - 1] = '\0';
		errmsg = pom;
	}

	gui_err(errmsg);
}

void gui_strerror(const gchar *errmsg)
{
	if (errmsg == NULL)
	{
		gui_err(errmsg);
		return;
	}

	gui_errfmt("%s : %s", gettext(errmsg), strerror(errno));
}

#if GTK_MAJOR_VERSION > 1
static void gui_fname_error(const gchar *fname)
{
	gchar		*fn, *err;
	gchar		pom[kMessageBufferSize];

	if (fname == NULL)
	{
		msg_set(&error_msg, fname);
		return;
	}

	err = utf8_from_str(strerror(errno));
	fn  = utf8_from_filename(fname, NULL);

	g_snprintf(pom, sizeof(pom), "%s : %s",
		(fn == NULL ? "" : fn), (err == NULL ? "" : err));

	utf8_free(fn);
	utf8_free(err);

	msg_set(&error_msg, pom);
	gdk_beep();
}
#endif /* GTK_MAJOR_VERSION > 1 */

static void Quit(GtkWidget *w, gpointer data)
{
	if (gui_option[kOptionSaveConfigOnExit].active)
	{
		nscache_gui_started = FALSE;
		save_rc();
	}

	gtk_main_quit();
}

static void About(GtkWidget *w, gpointer data)
{
	static GtkWidget *about_win = NULL;

	gchar		*str, pom[kMessageBufferSize];
	GtkWidget	*button, *image, *frame, *hbox, *label, *vbox;

	GtkWidget *window = about_win;

	if (window)
	{
		GuiWindowPresent(window);
		return;
	}

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	about_win = window;

	gtk_container_set_border_width(GTK_CONTAINER(window), 4);
	GuiWindowSetResizable(GTK_WINDOW(window), FALSE);
	GuiSignalConnect(window, "destroy", gtk_widget_destroyed, &about_win);

	str = guitl_gettext_utf8_from_str(gettext_nop("MozCache: About"));
	gtk_window_set_title(GTK_WINDOW(window), str);
	utf8_free(str);

	vbox = gtk_vbox_new(FALSE, 1);
	gtk_container_add(GTK_CONTAINER(window), vbox);

	hbox = gtk_hbox_new(FALSE, 1);
	frame = guitl_frame_new_add_child(NULL, hbox);
	gtk_box_pack_start(GTK_BOX(vbox), frame, TRUE, FALSE, 10);

	image = guitl_get_icon_dialog(kIconDialogNsCacheLogo);
	gtk_box_pack_start(GTK_BOX(hbox), image, TRUE, FALSE, 5);

	g_snprintf(pom, sizeof(pom), "%s\n\n"
					"%s\n\n"
					"%s %s\n\n"
					"%s\n"
					"%s\n"
					"%s\n\n"
					"%s\n"
					"%s\n"
					"%s",
		gettext("Cache viewer for Mozilla, Netscape, Firefox & SeaMonkey browsers"),
		gettext("Distributed under GNU-GPL 2 or later"),
		gettext("Version"), nscache_version_str,
		gettext("Previous versions (NScache)"),
		gettext("Author: Stefan Ondrejicka <ondrej@idata.sk>"),
			"URL: http://www.idata.sk/~ondrej/nscache/",
		gettext("Modifications in later versions (MozCache)"),
		gettext("Author: Harald Foerster <harald_foerster@users.sourceforge.net>"),
			"URL: http://mozcache.sourceforge.net/");

	pom[sizeof(pom) - 1] = '\0';
	str = utf8_from_str(pom);
	label = gtk_label_new(str == NULL ? "???" : str);
	utf8_free(str);

	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(label), 0.1, 0.5);
	gtk_misc_set_padding(GTK_MISC(label), 15, 15);
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, FALSE, 5);

	hbox = guitl_hbutton_box_new(vbox, 8, GTK_BUTTONBOX_END);
	button = guitl_button_box_add_icon_button(hbox, kIconButtonClose);
	gtk_widget_grab_default(button);
	GuiSignalConnectSwapped(button, "clicked", gtk_widget_destroy, window);

	gtk_widget_show_all(window);

} /* static void About(GtkWidget *w, gpointer data) */

static void BrowseOK(GtkFileSelection *fs)
{
	gchar		*fname;
	const gchar	*str;
	GtkWidget	*browse_type;

	str   = gtk_file_selection_get_filename(fs);
	fname = utf8_from_filename(str, NULL);

	if (fname == NULL)
	{
		gui_err(gettext(invalid_filename_str));
		gtk_widget_destroy(GTK_WIDGET(fs));

		return;
	}

	gtk_entry_set_text(entry_db, fname);
	utf8_free(fname);

	browse_type = GuiObjectGetUserData(fs);
	nscache_db.type = GuiOptionMenuGetIndex(browse_type);

	gtk_widget_destroy(GTK_WIDGET(fs));

	Open(entry_db);
}

static void Browse(GtkWidget *w, gpointer data)
{
	gint		i;
	gchar		*str;
	const gchar	*fname;
	GtkWidget	*box, *browse_type, *label, *menu;
	GtkFileSelection *fs;

	str = guitl_gettext_utf8_from_str(gettext_nop("Select cache index ..."));
	fs = GTK_FILE_SELECTION( gtk_file_selection_new(str) );
	utf8_free(str);

	gtk_file_selection_hide_fileop_buttons(fs);
	gtk_window_set_modal(GTK_WINDOW(fs), TRUE);

	box = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_start(GTK_BOX(fs->main_vbox), box, FALSE, FALSE, 0);

	label = guitl_label_new_gettext_utf8_from_str(gettext_nop("Cache index file type: "));
	gtk_box_pack_start(GTK_BOX(box), label, FALSE, FALSE, 2);

	browse_type = gtk_option_menu_new();
	GuiObjectSetUserData(fs, browse_type);

#ifndef MULTI_DB_TYPES
	gtk_widget_set_sensitive(browse_type, FALSE);
#endif
	menu	= gtk_menu_new();
	i	= NSCACHE_RT_FIRST;

	do
	{
		GtkWidget *mi;

		mi = guitl_get_icon_menu(i);
		GuiMenuAppend(menu, mi);

#ifdef HAVE_MSIE
		if (i == NSCACHE_RT_MSIE)
		{
			gtk_widget_set_sensitive(mi, FALSE);
		}
#endif
		GuiOptionMenuSetIndex(mi, i);
	}
	while (++i <= NSCACHE_RT_LAST);

	gtk_option_menu_set_menu(GTK_OPTION_MENU(browse_type), menu);

#ifdef HAVE_MSIE
	gtk_option_menu_set_history(GTK_OPTION_MENU(browse_type),
					(nscache_db.type == NSCACHE_RT_MSIE ?
							NSCACHE_RT_FIRST : nscache_db.type));
#else
	gtk_option_menu_set_history(GTK_OPTION_MENU(browse_type), nscache_db.type);
#endif
	gtk_box_pack_start(GTK_BOX(box), browse_type, FALSE, FALSE, 2);

	GuiSignalConnectSwapped(fs->cancel_button, "clicked", gtk_widget_destroy, fs);
	GuiSignalConnectSwapped(fs->ok_button, "clicked", BrowseOK, fs);

	fname = gtk_entry_get_text(entry_db);

	if (fname && *fname)
	{
		str = utf8_to_filename(fname);

		if (str != NULL)
		{
			gtk_file_selection_set_filename(fs, str);
			utf8_free(str);
		}
	}

	gtk_widget_show_all(GTK_WIDGET(fs));

} /* static void Browse(GtkWidget *w, gpointer data) */

static void Save(GtkWidget *ok_btn, nscache_record_t *rec)
{
	static gchar* err_str[] =
	{
		NULL,
		gettext_nop("Error opening source file"),
		gettext_nop("Error opening destination file"),
		gettext_nop("Error writing to destinantion file"),
		gettext_nop("Error linking files")
	};

	GtkWidget	*fs;
	const gchar	*dst;
	gint		error;

	fs	= gtk_widget_get_toplevel(ok_btn);
	dst	= gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs));
	error	= file_save(rec, (char *) dst);

	if (error == 0)
	{
		gui_msg(gettext("File copied"));
	}

	/* abs(error) */
	gui_strerror(err_str[-error]);

	gtk_widget_destroy(fs);
}

#ifdef HAVE_LINK
static void gui_hardlink_callback(GtkWidget *toggle_button)
{
	hardlink = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(toggle_button));
	gprop_set_bool(kPropBoolSaveHardlink, hardlink);
}
#endif

static void gui_single_folder_callback(GtkWidget *toggle_button)
{
	single_folder = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(toggle_button));
	gprop_set_bool(kPropBoolSaveSingleFolder, single_folder);
}

static void gui_url_to_local_callback(GtkWidget *option_menu)
{
	url_to_local_selector = GuiOptionMenuGetIndex(option_menu);
	gprop_set_int(kPropIntSaveUrlToLocal, url_to_local_selector);
}

static void gui_filename_callback(GtkFileSelection *fs)
{
	gchar *path, *sep;

	path = g_strdup(gtk_file_selection_get_filename(fs));

	while ((sep = strrchr(path, '/')) != NULL)
	{
		struct stat dirstat;

		sep[1] = '\0';

		if ( stat(path, &dirstat) == 0 &&
					S_ISDIR(dirstat.st_mode) )
		{
			gprop_set_str(kPropStrSavePath, path);
			break;
		}

		*sep = '\0';
	}

	g_free(path);
}

static gchar *gui_save_get_path(const gchar *fname)
{
	gchar *path, *sep;

	if (gprop_get_str(kPropStrSavePath, &path) == FALSE)
	{
		path = "";
	}

	sep = strrchr(path, '/');

	sep = (sep == NULL || sep[1] == '\0') ? "" : "/";

	return g_strconcat(path, sep, (fname == NULL) ? "noname" : fname, NULL);
}

static gchar *SaveFileDialog(GtkFileSelection *fs, GtkWidget *button, ListSelection *list)
{
	const gchar	*fname;
	gchar		*path, *str;

#ifdef HAVE_LINK
	/* check button - hardlink */
	gtk_box_pack_start(GTK_BOX(fs->main_vbox), button, FALSE, FALSE, 0);
#endif
	GuiSignalConnect(fs->ok_button, "clicked", Save, list->nscrec);

	str = strstr(list->nscrec->urlstr, "://");

	if (str)
	{
		fname = str + 3;
		str   = strstr(fname, "://");

		if (str)
		{
			/* URL in URL */

			unsigned int colon;

			colon = str - fname;
			fname = g_strdup(fname);
			str   = (gchar *) &(fname[colon]);

			while ((str = strchr(++str, '/')) != NULL)
			{
				/* replace '/' with '-' */
				*str = '-';
			}

			str = (gchar *) fname;
		}

		fname = strrchr(fname, '/');

		if (fname == NULL || *(++fname) == '\0')
		{
			fname = &index_html_str[1];
		}
	}

	else
	{
		fname = &index_html_str[1];
	}

	path = gui_save_get_path(fname);

	g_free(str);

	return path;

} /* static gchar *SaveFileDialog(GtkFileSelection *fs, GtkWidget *button, ListSelection *list) */

static void collect_nscache_records(GuiTreeStore *tree, nscache_record_t* rec,
						nscache_record_t*** rec_array)
{
	if (rec)
	{
		**rec_array = rec;
		(*rec_array)++;
	}
}

static void SaveFolder(GtkWidget *ok_btn, ListSelection *list)
{
	static const char fail_mesg[] = gettext_nop("Cannot create folder");

	size_t			rec_items;
	nscache_record_t**	rec_array;
	nscache_record_t**	rec_malloc;

	union
	{
		struct node_info  nfo;
		char		  mesg[4096];
		nscache_record_t* recs[1024];
	} u;

	/* file selector */
	GtkWidget *fs = gtk_widget_get_toplevel(ok_btn);

	rec_array  = u.recs;
	rec_malloc = NULL;

	/* number of files plus terminating 'NULL' */
	compute_node_info(list, &u.nfo);
	rec_items = u.nfo.filenum + 1;

	if ( (rec_items <= sizeof(u.recs) / sizeof(nscache_record_t*)) ||
		(rec_malloc = rec_array =
			malloc(rec_items * sizeof(nscache_record_t*))) != NULL )
	{
		static const char saved_fmt_str[] = "%s: %d   %s: %d   %s: %d   %s: %d";

		int		error;
		const char	*str;
		SaveDirResult	sdr;

		nscache_record_t** array_start = rec_array;

		/* fill array with pointers ... */
		guitl_tree_recursive(&(list->store.tree), TRUE,
			(GuiTreeStoreRecursiveFunc) collect_nscache_records, &rec_array);

		/* ... and terminate it with 'NULL' */
		*rec_array = NULL;

		/* write folder to this destination */
		str = gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs));

		error = folder_save(str,
			(const nscache_record_t**) array_start,
			GuiTreeGetLevel(&(list->store.tree)) - kTreeLevelDomain, &sdr);

		if (url_to_local_selector == URL_CONVERT_NONE)
		{
			g_snprintf(u.mesg, sizeof(u.mesg), &saved_fmt_str[9],
				gettext("folders created"), sdr.folders,
				gettext("files copied"), sdr.files,
				gettext("errors"), sdr.errors);
		}

		else
		{
			g_snprintf(u.mesg, sizeof(u.mesg), saved_fmt_str,
				gettext("folders created"), sdr.folders,
				gettext("files copied"), sdr.files,
				gettext("URLs converted"), sdr.converted,
				gettext("errors"), sdr.errors);
		}

		u.mesg[sizeof(u.mesg) - 1] = '\0';
		gui_msg(u.mesg);

		if (error == 0)
		{
			str = sdr.fname;

			if (str)
			{
				errno = sdr.errnr;
			}

			gui_fname_error(str);
		}

		else
		{
			gui_strerror(fail_mesg);
		}

		if (sdr.fname)
		{
			free(sdr.fname);
		}

	} /* if( (rec_items <= ... */

	else
	{
		gui_strerror(fail_mesg);
	}


	if (rec_malloc)
	{
		free(rec_malloc);
	}


	gtk_widget_destroy(fs);

} /* static void SaveFolder(GtkWidget *ok_btn, ListSelection *selected) */

static gchar *SaveFolderDialog(GtkFileSelection *fs, GtkWidget *button, ListSelection *list)
{
	GtkWidget	*label, *menu, *option, *frame, *widget;
	GtkTable	*table;
	gchar		*str;
	gint		index;

	widget = gtk_table_new(2, 2, FALSE);
	frame = guitl_frame_new_add_child(NULL, widget);
	gtk_box_pack_start(GTK_BOX(fs->main_vbox), frame, FALSE, FALSE, 0);

	table = GTK_TABLE(widget);

	/* single_folder + hardlink */
	
#ifdef HAVE_LINK
	/* check button - hardlink */
	gtk_table_attach(table, button, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 4, 4);
#endif

	/* check button - single_folder */

	str = guitl_gettext_utf8_from_str(gettext_nop("Single Folder"));
	button = gtk_check_button_new_with_label(str);
	utf8_free(str);

	gtk_table_attach(table, button, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 4, 4);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), single_folder);
	GuiSignalConnectSwapped(fs->ok_button, "clicked", gui_single_folder_callback, button);

	/* convert URLs */

	label = guitl_label_new_gettext_utf8_from_str(gettext_nop("Convert URLs to local:"));
	gtk_table_attach(table, label, 1, 2, 0, 1, GTK_FILL, GTK_FILL, 4, 4);

	option = gtk_option_menu_new();
	gtk_table_attach(table, option, 1, 2, 1, 2, GTK_FILL, GTK_FILL, 4, 4);

	GuiSignalConnectSwapped(fs->ok_button, "clicked", gui_url_to_local_callback, option);
	GuiSignalConnect(fs->ok_button, "clicked", SaveFolder, list);

	menu   = gtk_menu_new();
	gtk_widget_realize(menu);

	index = URL_CONVERT_NONE;

	do
	{
		static const char *url_to_local_str[] =
		{
			gettext_nop("none"),
			gettext_nop("all URLs with domain reference"),
			gettext_nop("only URLs with cache file reference")
		};

		GtkWidget *mi;

		str = guitl_gettext_utf8_from_str(url_to_local_str[index]);
		mi = gtk_menu_item_new_with_label(str);
		utf8_free(str);

		GuiMenuAppend(menu, mi);
		GuiOptionMenuSetIndex(mi, index);

		gtk_widget_show(mi);
	}
	while (++index <= URL_CONVERT_ONLY_FILE_REFERENCE);

	/* option */

	gtk_option_menu_set_menu(GTK_OPTION_MENU(option), menu);
	gtk_option_menu_set_history(GTK_OPTION_MENU(option),
		(url_to_local_selector > URL_CONVERT_ONLY_FILE_REFERENCE) ?
					URL_CONVERT_NONE : url_to_local_selector);

	return gui_save_get_path(list->name);

} /* static gchar *SaveFolderDialog(GtkFileSelection*, GtkWidget*, ListSelection*) */

static void SaveAs(GtkWidget *w, ListSelection **selected)
{
	static const gchar save_file_str[]   = gettext_nop("Save cache file as ...");
	static const gchar save_folder_str[] = gettext_nop("Save cache folder as ...");

	gboolean	active;
	gchar		*str;
	const gchar	*title;
	GtkWidget	*button;
	GtkFileSelection *fs;

	ListSelection *list = *selected;

	if (list == NULL || (list->nscrec == NULL && list->type != kTreeList))
	{
		return;
	}

	title = (list->nscrec == NULL) ? save_folder_str : save_file_str;

	str = guitl_gettext_utf8_from_str(title);
	fs = GTK_FILE_SELECTION( gtk_file_selection_new(str) );
	utf8_free(str);

	gtk_file_selection_hide_fileop_buttons(fs);
	gtk_window_set_modal(GTK_WINDOW(fs), TRUE);

	GuiSignalConnectSwapped(fs->ok_button, "clicked", gui_filename_callback, fs);
	GuiSignalConnectSwapped(fs->cancel_button, "clicked", gtk_widget_destroy, fs);
	
#ifdef HAVE_LINK

	/* check button - hardlink */

	str = guitl_gettext_utf8_from_str(gettext_nop("Hard Link"));
	button = gtk_check_button_new_with_label(str);
	utf8_free(str);

	if (hardlink_enable == FALSE)
	{
		gtk_widget_set_sensitive(button, FALSE);
		active = FALSE;
	}

	else
	{
		GuiSignalConnectSwapped(fs->ok_button, "clicked", gui_hardlink_callback, button);
		active = hardlink;
	}

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), active);
#endif
	if (list->nscrec == NULL)
	{
		str = SaveFolderDialog(fs, button, list);
	}

	else
	{
		str = SaveFileDialog(fs, button, list);
	}

	gtk_file_selection_set_filename(fs, str);
	g_free(str);

	gtk_widget_show_all(GTK_WIDGET(fs));

} /* static void SaveAs(GtkWidget *w, ListSelection **selected) */

#ifndef CACHE_READONLY

static GSList *gui_tree_append_parents(GSList *node_list, GuiTreeStore *tree)
{
	GSList *last;

#ifdef CONFIG_GTKCLIST_GTKCTREE

	GtkCTreeNode	*parent;
	GtkCTreeRow	*row;

	last = g_slist_last(node_list);
	row  = GTK_CTREE_ROW(last->data);

	while ((parent = row->parent) != NULL)
	{
		node_list = g_slist_append(node_list, parent);
		row	  = GTK_CTREE_ROW(parent);
	}
#else
 	GtkTreeIter *child, *parent;

	/* the selected node */
	last   = g_slist_last(node_list);
	child  = (GtkTreeIter *) last->data;

	/* all parent nodes till reaching toplevel */
	while (parent = g_malloc(sizeof(GtkTreeIter)),
		gtk_tree_model_iter_parent(GTK_TREE_MODEL(tree->stree),
								parent, child))
	{
		node_list = g_slist_append(node_list, parent);
		child     = parent;
	}

	g_free(parent);
#endif
	return node_list;

} /* static GSList *gui_tree_append_parents(GSList *node_list, GuiTreeStore *tree) */

#ifndef CONFIG_GTKCLIST_GTKCTREE

static void gui_store_remove_node(ListSelection *list, GtkTreeIter *iter)
{
	GtkTreeModel		*model;
	GtkTreeSelection	*selection;
	GtkTreeIter		*parent;
	GtkTreeIter		parent_iter;

	if (list->type == kTreeList)
	{
		model = GTK_TREE_MODEL(list->store.tree.stree);
		selection = list->store.tree.selection;
	}

	else
	{
		model = GTK_TREE_MODEL(list->store.list.slist);
		selection = list->store.list.selection;
	}

	 parent = &parent_iter;
	*parent = *iter;

	if (gtk_tree_model_iter_has_child(model, parent) == FALSE)
	{
		GtkTreeIter next_iter;

		GtkTreeIter *next = NULL;

		if (gtk_tree_selection_iter_is_selected(selection, parent))
		{
			for (;;)
			{
				gint num;

				next_iter = *parent;

				if (gtk_tree_model_iter_next(model, &next_iter))
				{
					next = &next_iter;
					break;
				}

				next_iter = *parent;

				if (gtk_tree_model_iter_parent(model, parent,
								&next_iter) == FALSE)
				{
					parent = NULL;
				}

				num = gtk_tree_model_iter_n_children(model, parent);

				if (num > 1)
				{
					if (gtk_tree_model_iter_nth_child(model,
							&next_iter, parent, num - 2))
					{
						next = &next_iter;
					}

					break;
				}

				if (parent == NULL)
				{
					break;
				}

			} /* for (;;) */

		} /* if (gtk_tree_selection_iter_is_selected(selection, parent)) */

		if (list->type == kTreeList)
		{
			gtk_tree_store_remove(list->store.tree.stree, iter);
		}

		else
		{
			gtk_list_store_remove(list->store.list.slist, iter);
		}

		if (next)
		{
			gtk_tree_selection_select_iter(selection, next);
		}

	} /* if (gtk_tree_model_iter_has_child(model, parent) == FALSE) */

} /* static void gui_select_next_iter(GtkTreeModel*, GtkTreeSelection*, GtkTreeIter*) */

#endif /* #ifndef CONFIG_GTKCLIST_GTKCTREE */

static void gui_list_remove_node(ListSelection *list, nscache_record_t *rec)
{
	if (rec && list->type == kSortedList)
	{
#ifdef CONFIG_GTKCLIST_GTKCTREE

		gint row = guitl_list_find_by_row_data(&(list->store.list), rec);

		if (row >= 0)
		{
			gtk_clist_remove(list->store.list.clist, row);
		}

#else /* ! CONFIG_GTKCLIST_GTKCTREE */

		GtkTreeIter *iter = guitl_list_find_by_row_data(&(list->store.list), rec);

		if (iter)
		{
			gui_store_remove_node(list, iter);
			g_free(iter);
		}

#endif/* #ifndef CONFIG_GTKCLIST_GTKCTREE */

	} /* if (rec && list->type == kSortedList) */

} /* static void gui_list_remove_node(ListSelection *list, gpointer rec) */

static void gui_tree_remove_node(gpointer node, ListSelection *list)
{

#ifdef CONFIG_GTKCLIST_GTKCTREE

	if (GTK_CTREE_ROW(node)->children == NULL)
	{
		gtk_ctree_remove_node(list->store.tree.ctree, node);
	}
#else
	gui_store_remove_node(list, node);
	g_free(node);
#endif

} /* static void gui_tree_remove_node(gpointer node, gpointer tree) */

static void gui_remove_nscache_record_recursive(GuiTreeStore *tree,
				nscache_record_t* rec, GSList **node_list)
{
	if (rec)
	{
		if (nscache_record_delete(rec) != 0)
		{
			return;
		}

		g_slist_foreach(list_selection_chain,
			(GFunc) gui_list_remove_node, rec);
	}

#ifdef CONFIG_GTKCLIST_GTKCTREE
	*node_list = g_slist_prepend(*node_list, tree->node);
#else
	{
		GtkTreeIter *iter = g_malloc(sizeof(GtkTreeIter));
		*iter = tree->iter;

		*node_list = g_slist_prepend(*node_list, iter);
	}
#endif

} /* static void gui_remove_nscache_record_recursive(GuiTreeStore*,
					nscache_record_t*, GSList**) */

static void Delete(GtkWidget *ok_btn, ListSelection *list)
{
	GuiTreeStore	*tree;
	GSList		*node_list;

	GtkWidget *dialog = gtk_widget_get_toplevel(ok_btn);
	gtk_widget_destroy(dialog);

#ifdef CONFIG_GTKCLIST_GTKCTREE
	g_slist_foreach(list_selection_chain,
		(GFunc) gui_clist_freeze_or_thaw, (gpointer) gtk_clist_freeze);
#endif
	tree      = &(notebook_list[kTreeList].store.tree);
	node_list = NULL;

	if (list->nscrec)
	{
		if (nscache_record_delete(list->nscrec) == 0)
		{
			gpointer node;
			node = guitl_tree_find_by_row_data(tree, list->nscrec);

			if (node)
			{
				node_list = g_slist_append(node_list, node);
			}

			g_slist_foreach(list_selection_chain,
				(GFunc) gui_list_remove_node, list->nscrec);
		}
	}

	else if (list->type == kTreeList)
	{
		guitl_tree_recursive(tree, TRUE, (GuiTreeStoreRecursiveFunc)
				gui_remove_nscache_record_recursive, &node_list);
	}

	if (node_list)
	{
		node_list = gui_tree_append_parents(node_list, tree);
		g_slist_foreach(node_list, (GFunc) gui_tree_remove_node,
							&notebook_list[kTreeList]);
		g_slist_free(node_list);

		g_slist_foreach(search_dialog_list,
			(GFunc) search_update_window_title, NULL);

		update_entry_info(list);
		print_node_info(NULL);
	}

#ifdef CONFIG_GTKCLIST_GTKCTREE
	g_slist_foreach(list_selection_chain,
		(GFunc) gui_clist_freeze_or_thaw, (gpointer) gtk_clist_thaw);
#endif

} /* static void Delete(GtkWidget *button, ListSelection *list) */

static GtkWidget *delete_dialog_get_label(ListSelection *list)
{
	static const gchar rm_file[] =
		gettext_nop("You are in progress to delete the following file:");

	static const gchar rm_folder[] =
		gettext_nop("You are in progress to delete the following folder:");

	GtkWidget	*label;
	const gchar	*what;
	gchar		*str, *name;
	gchar		pom[kMessageBufferSize], buf[64];

	name = list->name;

	if (list->nscrec == NULL)
	{
		what = rm_folder;
	}

	else
	{
		what = rm_file;

		if (list->type == kSortedList)
		{
			name = strrchr(name, '/');
		}
	}

	if (name == NULL || (*(name) == '/' && *(++name) == '\0'))
	{
		name = (gchar *) &index_html_str[1];
	}

	if (strlen(name) > 32)
	{
		name = strncpy(buf, name, 27);
		strcpy(&buf[27], "(...)");
	}

	g_snprintf(pom, sizeof(pom), "%s\n<<< %s >>>\n%s",
			gettext(what), name, gettext("Do you like to continue?"));

	str = utf8_from_str(pom);
	label = gtk_label_new(str == NULL ? name : str);
	utf8_free(str);

	return label;
}

static void DeleteDialog(GtkWidget *w, ListSelection **selected)
{
	GtkWidget *button, *dialog, *hbox, *image, *label;

	ListSelection *list = *selected;

	if(list == NULL || list->name == NULL)
	{
		return;
	}

	dialog = gtk_dialog_new();
	gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
	GuiWindowSetResizable(GTK_WINDOW(dialog), FALSE);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), hbox);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 10);

	image = guitl_get_icon_dialog(kIconDialogWarning);
	gtk_container_add(GTK_CONTAINER(hbox), image);

	label = delete_dialog_get_label(list);
	gtk_container_add(GTK_CONTAINER(hbox), label);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
	GuiWidgetSetSizeRequest(label, -1, 75);

	hbox = guitl_hbutton_box_new(GTK_DIALOG(dialog)->action_area,
							8, GTK_BUTTONBOX_END);

	button = guitl_button_box_add_icon_button(hbox, kIconButtonCancel);
	GuiSignalConnectSwapped(button, "clicked", gtk_widget_destroy, dialog);
	gtk_widget_grab_default(button);

	button = guitl_button_box_add_icon_button(hbox, kIconButtonDelete);
	GuiSignalConnect(button, "clicked", Delete, windata.toplist);

	gtk_widget_show_all(dialog);

} /* static void DeleteDialog(GtkWidget *w, ListSelection **selected) */

#endif /* ! CACHE_READONLY */

static GtkWidget *gui_properties_fill_frame(char *title, char *str)
{
	GtkWidget *label;

	str = utf8_from_str(str);
	label = gtk_label_new(str);
	utf8_free(str);

	gtk_misc_set_alignment(GTK_MISC(label), 0.5, 0.5);
	gtk_misc_set_padding(GTK_MISC(label), 8, 8);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);

	return guitl_frame_new_add_child(title, label);
}

static void PropertiesDialog(GtkWidget *w, gpointer data)
{
	gint		icon_idx, len;
	GtkWidget	*button, *dlg, *frame, *hbox, *image, *vbox;
	const gchar	*noav_str;
	gchar		*str;
	CacheDB		*cache_db;
	gchar		pom[kMessageBufferSize], timebuf[30];

	union
	{
		nscache_record_t	rec;
		struct stat		pstat;
	} u;

	dlg = gtk_dialog_new();

	gtk_window_set_default_size(GTK_WINDOW(dlg), -1, -1);
	gtk_window_set_modal(GTK_WINDOW(dlg), TRUE);
	GuiWindowSetResizable(GTK_WINDOW(dlg), FALSE);

	str = guitl_gettext_utf8_from_str(gettext_nop("MozCache: Properties"));
	gtk_window_set_title(GTK_WINDOW(dlg), str);
	utf8_free(str);

	hbox = guitl_hbutton_box_new(GTK_DIALOG(dlg)->action_area, 8, GTK_BUTTONBOX_END);
	button = guitl_button_box_add_icon_button(hbox, kIconButtonClose);
	GuiSignalConnectSwapped(button, "clicked", gtk_widget_destroy, dlg);
	gtk_widget_grab_default(button);

	noav_str = gettext(na_str);
	icon_idx = kIconDialogError;
	cache_db = &nscache_db;

	if (cache_db->name == NULL)
	{
		len = g_snprintf(pom, sizeof(pom), "%s",
			gettext("Cache Index File not specified!"));
	}

	else
	{
		len = g_snprintf(pom, sizeof(pom), "%s\n%s: %s",
				cache_db->name, gettext("Version"),
					(cache_db->version == NULL ?
						noav_str : cache_db->version));

		if (len > 0 && len < sizeof(pom))
		{
#ifdef HAVE_MSIE
		if (nscache_db.type == NSCACHE_RT_MSIE)
		{
			icon_idx = kIconDialogInfo;
		}
		else
		{
#endif
			if (u.rec.filename = cache_db->name,
				u.rec.content_offset = 0,
					file_status(&(u.rec), &(u.pstat)) != 0)
			{
				len += g_snprintf(&pom[len], sizeof(pom) - len, "\n%s\n%s",
					gettext("Can't get status of Cache Index File"),
								strerror(errno));
			}

			else
			{
				gint		index_acc;
				const gchar	*byte_ptr, *srd, *swr;

				unsigned long size = u.pstat.st_size;

				if (size >= 1024)
				{
					size /= 1024;
					byte_ptr = gettext("kB");
				}

				else
				{
					byte_ptr = gettext("Bytes");
				}

				index_acc = file_access_index();

				if (index_acc < 0)
				{
					srd = swr = noav_str;
				}

				else
				{
					srd = swr = gettext(yes_str);

					if ( ! (index_acc & FILE_ACCESS_READ))
					{
						srd = gettext(no_str);
					}

					if ( ! (index_acc & FILE_ACCESS_WRITE))
					{
						swr = gettext(no_str);
					}
				}

				str = tl_time_to_str(timebuf, sizeof(timebuf),
								&(u.pstat.st_mtime));
				if (str == NULL)
				{
					str = (gchar *) noav_str;
				}

				len += g_snprintf(&pom[len], sizeof(pom) - len,
							"\n%s: %lu %s\n"
							"%s: %s\n"
							"%s: %s\n"
							"%s: %s",
					gettext("Size"), size, byte_ptr,
					gettext("Readable"), srd,
					gettext("Writable"), swr,
					gettext("Modification time"), str);

				icon_idx = kIconDialogInfo;
			}
#ifdef HAVE_MSIE
		} /* if (nscache_db.type == NSCACHE_RT_MSIE) .. else */
#endif
		} /* if (len > 0 && len < sizeof(pom)) */
	}

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dlg)->vbox), hbox);
	gtk_container_set_border_width(GTK_CONTAINER(hbox), 10);

	image = guitl_get_icon_dialog(icon_idx);
	gtk_container_add(GTK_CONTAINER(hbox), image);

	vbox = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(hbox), vbox);

	str = pom;
	str[sizeof(pom) - 1] = '\0';
	frame = gui_properties_fill_frame(gettext_nop("Cache Index File"), str);
	gtk_container_add(GTK_CONTAINER(vbox), frame);

	len = get_node_info(pom, sizeof(pom), kNodeInfoFormatMultiLine, NULL);

	if (len > 0 && len < sizeof(pom) && cache_db->name)
	{
		str = tl_time_to_str(timebuf, sizeof(timebuf),
						&(cache_db->access_time));
		if (str == NULL)
		{
			str = (gchar *) noav_str;
		}

		g_snprintf(&pom[len], sizeof(pom) - len,
				"\n%s: %s", gettext("Access time"), str);
	}

	str = pom;
	str[sizeof(pom) - 1] = '\0';
	frame = gui_properties_fill_frame(gettext_nop("Contents"), str);
	gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 10);

	gtk_widget_show_all(dlg);
}

static void NodeInfo(GtkWidget *w, ListSelection **selection)
{
	print_node_info(*selection);
}

static void chmod_recursive(GuiTreeStore *tree, nscache_record_t *rec, gpointer rdwr)
{
	if (rec && rec->content_offset == 0)
	{
		struct stat estat;

		if (file_status(rec, &estat) == 0)
		{
			if (rdwr)
			{
				estat.st_mode |= S_IWUSR;
			}

			else
			{
				estat.st_mode &= ~S_IWUSR;
			}

			estat.st_mode |= S_IRUSR;

			if (file_chmod(rec, estat.st_mode))
			{
				gui_err(gettext("Can't change mode for any of cache files"));
			}
		}

		else
		{
			gui_err(gettext("Some of cache file doesn't exist"));
		}
	}
}

static void RdOnly(GtkWidget *w, ListSelection **selected)
{
	ListSelection *list = *selected;

	if (list)
	{
		if (list->nscrec == NULL && list->type == kTreeList)
		{
			/* folder */
			guitl_tree_recursive(&(list->store.tree), TRUE,
				(GuiTreeStoreRecursiveFunc) chmod_recursive,
								(gpointer) FALSE);
		}

		else
		{
			/* single file */
			chmod_recursive(NULL, list->nscrec, (gpointer) FALSE);
		}
	}
}

static void Writable(GtkWidget *w, ListSelection **selected)
{
	ListSelection *list = *selected;

	if (list)
	{
		if (list->nscrec == NULL && list->type == kTreeList)
		{
			/* folder */
			guitl_tree_recursive(&(list->store.tree), TRUE,
				(GuiTreeStoreRecursiveFunc) chmod_recursive,
								(gpointer) TRUE);
		}

		else
		{
			/* single file */
			chmod_recursive(NULL, list->nscrec, (gpointer) TRUE);
		}
	}
}

static gint watch_viewer_status(gpointer data)
{
	struct
	{
		guint	tid;
		gint	limit;
		gchar*	appassign;

	} *viewer_timeout = data;

	if (viewer_status)
	{
		gui_errfmt("%s : %s", viewer_timeout->appassign, strerror(viewer_status));
		viewer_timeout->limit = 0;
	}

	if (--viewer_timeout->limit <= 0)
	{
		g_free(viewer_timeout->appassign);
		viewer_timeout->appassign = NULL;
		viewer_timeout->tid = 0;

		return FALSE;
	}

	return TRUE;
}

static void _start_viewer(const gchar *appassign, gint viewer_type)
{
	static struct
	{
		guint	tid;
		gint	limit;
		gchar*	appassign;

	} viewer_timeout = {0, 0, NULL};

	if (viewer_start(appassign, (windata.toplist)->nscrec, viewer_type))
	{
		static gchar *viewer_error_msg[] =
		{
			gettext_nop("URL viewer can't be launched"),
			gettext_nop("File viewer can't be launched")
		};

		gui_strerror(viewer_error_msg[viewer_type]);
	}

	else
	{
		static gchar *viewer_launched_msg[] =
		{
			gettext_nop("URL viewer launched"),
			gettext_nop("File viewer launched")
		};

		/*
			The viewer was successfully started. Now
			we have a look at the status every 200 ms
			25 times - maybe the viewer will finish
			with an error, and we can inform the user
			with a message - if so.
		*/

		viewer_timeout.limit = 25;

		if(viewer_timeout.tid)
		{
			GuiTimeoutRemove(viewer_timeout.tid);
			g_free(viewer_timeout.appassign);
		}

		viewer_timeout.appassign = g_strdup(appassign);

		viewer_timeout.tid = GuiTimeoutAdd(200,
				watch_viewer_status, &viewer_timeout);

		gui_msg(gettext(viewer_launched_msg[viewer_type]));
		gui_err(NULL);
	}
}

static void ViewFile(GtkWidget *w, const gchar *args_file)
{
	_start_viewer(args_file, VIEWER_FILE);
}

static void ViewURL(GtkWidget *w, const gchar *args_url)
{
	_start_viewer(args_url, VIEWER_URL);
}

static gint ViewerDefault(gint viewer_type)
{
	ListSelection *list = windata.toplist;

	if (list && list->nscrec)
	{
		GSList *ptr;

		ptr = apassign_get_matching_viewers(list->nscrec->content_type);

		while (ptr)
		{
			const gchar	*app;
			apassign	*e;

			e = ptr->data;

			if (viewer_type == VIEWER_FILE)
			{
				int cache_acc = file_access(list->nscrec);

				if (cache_acc <= 0 || ! (cache_acc & FILE_ACCESS_READ))
				{
					return TRUE;
				}

				app = e->args_file;
			}

			else
			{
				app = e->args_url;
			}

			if (app)
			{
				_start_viewer(app, viewer_type);
				break;
			}

			ptr = ptr->next;
		}

		g_slist_free(ptr);
	}

	return TRUE;
}

static void GenViewerMenu(GtkWidget *submenu, int viewer_type)
{
	GSList *ptr, *mlabels;

	ListSelection *list = windata.toplist;

	if (list == NULL || list->nscrec == NULL)
	{
		return;
	}

	while (GTK_MENU_SHELL(submenu)->children)
	{
		gtk_widget_destroy(GTK_WIDGET(GTK_MENU_SHELL(submenu)->children->data));
	}

	mlabels = NULL;

	ptr = apassign_get_matching_viewers(list->nscrec->content_type);

	while (ptr)
	{
		apassign *e = ptr->data;
		gchar *app;

		app = (viewer_type == VIEWER_FILE) ?
					e->args_file : e->args_url;
		if (app)
		{
			gchar *s, *str;

			str = strrchr(app, ' ');

			if ((s = str) != NULL)
			{
				while (*(++s) == ' ') {;}

				if (strcmp(s, uof_param_str[viewer_type]))
				{
					str = NULL;
				}
			}

			if (str > app)
			{
				str = g_strndup(app, str - app);
			}

			else
			{
				str = g_strdup(app);
			}

			if (g_slist_find_custom(mlabels, str,
					(GCompareFunc) strcmp) == NULL)
			{
				GtkWidget *mi;

				mlabels = g_slist_append(mlabels, str);

				str = utf8_from_str(str);

				if (str)
				{
					mi = gtk_menu_item_new_with_label(str);
					utf8_free(str);

					GuiMenuAppend(submenu, mi);
					gtk_widget_show(mi);

					GuiSignalConnect(mi, "activate",
						(viewer_type == VIEWER_FILE ?
							ViewFile : ViewURL), app);
				}
			}
		}

		ptr = g_slist_remove(ptr, ptr->data);
	}

	g_slist_foreach(mlabels, (GFunc) g_free, NULL);
	g_slist_free(mlabels);
}

static void SelectionSetURL(GtkWidget *w, ListSelection **selected)
{
	ListSelection *list = *selected;

	if (list && list->nscrec && list->name)
	{
		g_free(selection_urlstr);

		if (gtk_selection_owner_set(windata.mainwin,
				GDK_SELECTION_PRIMARY, GDK_CURRENT_TIME))
		{
			selection_urlstr = g_strdup(list->name);
		}

		else if (gdk_selection_owner_get(GDK_SELECTION_PRIMARY) ==
								windata.mainwin->window)
		{
			gtk_selection_owner_set(NULL, GDK_SELECTION_PRIMARY,
								GDK_CURRENT_TIME);
		}
	}
}

static gint SelectionClear(GtkWidget *w, GdkEventSelection *event, gchar **urlstr)
{
	g_free(*urlstr);
	*urlstr = NULL;

	return FALSE;
}

static void SelectionSend(GtkWidget *w, GtkSelectionData *selection_data,
						guint info, guint time, gchar **urlstr)
{
	GdkAtom type;

	if (info >= SEL_TYPE_FIRST && info <= SEL_TYPE_LAST)
	{
		type = gdk_atom_intern(targetlist[info - 1].target, FALSE);
	}

	else
	{
		type = GDK_NONE;
	}

	gtk_selection_data_set(selection_data, type, 8 * sizeof(gchar),
						*urlstr, strlen(*urlstr));
}

static void update_entry_info(ListSelection *list)
{
	GtkWidget	*ilabel;
	gint		len;
	gchar		*str;
	gchar		pom[kMessageBufferSize];

	nscache_record_t *rec;

	if (info_dialog == NULL)
	{
		return;
	}

	rec = (list == NULL) ? NULL : list->nscrec;

	if (rec)
	{
		const gchar	*noav_str, *path;
		gchar		mdtm[30], atm[30], exptm[30];

#ifdef HAVE_MSIE
		if (rec->content_offset == MSIE_CACHE)
		{
			path = MSIE_GET_PATHNAME(rec);
		}
		else
#endif
		path = nscache_db.path.str;
		noav_str = gettext(na_str);

		len += g_snprintf(pom, sizeof(pom), "%s: %s\n"
							"%s: %s\n"
							"%s: %s\n"
							"%s: %s\n"
							"%s: %d %s\n"
							"%s: %s\n"
							"%s: %s\n"
							"%s: %s\n"
							"%s: %s\n"
							"%s: %s\n",
			gettext("Source URL"), rec->urlstr,
			gettext("Local cache file"), rec->filename,
			gettext("Path"), path,
			gettext("MIME type"), (rec->content_type ?
					rec->content_type : noav_str),
			gettext("Size"), rec->content_length, gettext("Bytes"),
			gettext("Encoding"), (rec->content_encoding ?
					rec->content_encoding : noav_str),
			gettext("Charset"), (rec->charset ?
						rec->charset : noav_str),
			gettext("Modification time"),
			(tl_time_to_str(mdtm, sizeof(mdtm), &(rec->last_modified)) ?
									mdtm : noav_str),
			gettext("Access time"),
			(tl_time_to_str(atm, sizeof(atm), &(rec->last_accessed)) ?
									atm : noav_str),
			gettext("Expiration time"),
			(tl_time_to_str(exptm, sizeof(exptm), &(rec->expires)) ?
									exptm : noav_str));
	}

	else
	{
		gchar *folder_name = (list == NULL) ? NULL : list->name;

		if (folder_name == NULL || *folder_name == '\0')
		{
			folder_name = gettext("no entry");
		}

		len = g_snprintf(pom, sizeof(pom), "%s\n\n", folder_name);

		if (len < 0 || len >= sizeof(pom))
		{
			goto print_entry_info;
		}

		get_node_info(&pom[len], sizeof(pom) - len, kNodeInfoFormatMultiLine, list);
	}

print_entry_info:
	pom[sizeof(pom) - 1] = '\0';
	str = utf8_from_str(pom);
	ilabel = GuiObjectGetUserData(info_dialog);
	gtk_label_set_text(GTK_LABEL(ilabel), str);
	utf8_free(str);

} /* static void update_entry_info(ListSelection *list) */

static void ViewInfo(GtkWidget *w, ListSelection **selected)
{
	static const gchar info_title[] = gettext_nop("MozCache: entry info window");

	GtkWidget *button, *dlg, *hbox, *ilabel, *swin;

	dlg = info_dialog;

	if (dlg)
	{
		GuiWindowPresent(dlg);
		return;
	}

	dlg = gtk_dialog_new();
	info_dialog = dlg;
	gui_window_setup(dlg, info_title, kPropIntInfoWinWidth);
	GuiSignalConnect(dlg, "destroy", gtk_widget_destroyed, &info_dialog);

	hbox = guitl_hbutton_box_new(GTK_DIALOG(dlg)->action_area, 0, GTK_BUTTONBOX_END);
	button = guitl_button_box_add_icon_button(hbox, kIconButtonClose);
	GuiSignalConnectSwapped(button, "clicked", gtk_widget_destroy, dlg);
	gtk_widget_grab_default(button);

	swin = gtk_scrolled_window_new(NULL, NULL);
	gtk_container_set_border_width(GTK_CONTAINER(GTK_DIALOG(dlg)->vbox), 3);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dlg)->vbox), swin);
	GuiWidgetSetSizeRequest(swin, 200, 100);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(swin),
					GTK_POLICY_ALWAYS, GTK_POLICY_NEVER);
	ilabel = gtk_label_new(NULL);
	GuiObjectSetUserData(dlg, ilabel);
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(swin), ilabel);

	gtk_label_set_justify(GTK_LABEL(ilabel), GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment(GTK_MISC(ilabel), 0.0, 0.0);
	gtk_misc_set_padding(GTK_MISC(ilabel), 12, 12);

	update_entry_info(*selected);
	gtk_widget_show_all(dlg);

} /* static void ViewInfo(GtkWidget *w, ListSelection **selected) */

#ifdef CONFIG_GTKCLIST_GTKCTREE
static gint gui_tree_compare_func(GtkCTree *ctree,
				GtkCTreeRow *r1, GtkCTreeRow *r2)
{
	if (r1->children && r2->children == NULL)
	{
		/* r1 is folder */
		return -1;
	}

	if (r2->children && r1->children == NULL)
	{
		/* r2 is folder */
		return 1;
	}

	/* r1 and r2 are both of the same type */
	return strcmp(r1->row.cell->u.text, r2->row.cell->u.text);
}
#else
static gint gui_tree_compare_func(GtkTreeModel *model, GtkTreeIter *iter1,
					GtkTreeIter *iter2, gpointer userdata)
{
	gint		retv;
	gboolean	expander1, expander2;
	gchar		*name1, *name2;

	expander1 = gtk_tree_model_iter_has_child(model, iter1);
	expander2 = gtk_tree_model_iter_has_child(model, iter2);

	if (expander1 && expander2 == FALSE)
	{
		/* expander1 is folder */
		return -1;
	}

	if (expander2 && expander1 == FALSE)
	{
		/* expander2 is folder */
		return 1;
	}

	/* expander1 and expander2 are both of the same type */
	gtk_tree_model_get(model, iter1, kColumnURL, &name1, -1);
	gtk_tree_model_get(model, iter2, kColumnURL, &name2, -1);

	retv = strcmp(name1, name2);

	g_free(name1);
	g_free(name2);

	return retv;
}
#endif /* #ifdef CONFIG_GTKCLIST_GTKCTREE .. #else */

#ifdef CONFIG_GTKCLIST_GTKCTREE
static gint gui_list_compare_func(GtkCList *clist, GtkCListRow *r1, GtkCListRow *r2)
#else
static gint gui_list_compare_func(GtkTreeModel *model, GtkTreeIter *iter1,
					GtkTreeIter *iter2, gpointer userdata)
#endif
{
	nscache_record_t *rec1, *rec2;

#ifdef CONFIG_GTKCLIST_GTKCTREE
	rec1     = r1->data;
	rec2     = r2->data;
#else
	gtk_tree_model_get(model, iter1, kColumnRowData, &rec1, -1);
	gtk_tree_model_get(model, iter2, kColumnRowData, &rec2, -1);
#endif

	return nscache_compare_func(rec1, rec2);
}

static void gui_list_sort(GtkWidget *menu, gint *stype_ptr)
{
	if (stype_ptr)
	{
		gint sort_type = *stype_ptr;

		if (sort_type == nscache_sort_type)
		{
			sort_type ^= STYPE_REVERS_MASK;
			*stype_ptr = sort_type;
		}

		nscache_sort_type = sort_type;
		gprop_set_int(kPropIntSortType, sort_type);
	}

	GuiListColumnSort(&(notebook_list[kSortedList].store.list),
						0, gui_list_compare_func, NULL);
}

gpointer get_parent_node(char **dname, GHashTable *folder_node)
{
	gpointer	parent;
	size_t		colon;
	char		*beg, *end, *urldup, *urlstr, *urlurl;

	urlstr = tl_get_url_start(*dname);

	if (urlstr == NULL)
	{
		return NULL;
	}

	*dname = urlstr;
	beg    = strchr(urlstr, ':');
	colon  = beg - urlstr;
	beg   += 3;
	urlurl = NULL;
	urldup = NULL;

	/* remove multiple slashes, if any */

	while ((end = strstr(beg, "//")) != NULL)
	{
		if (end[-1] == ':')
		{
			/*
				If we've got something like this
				"http://www.foo.bar/load.html?http://www.foo.bar/index.html"
				then the file name will set to
				"load.html?http://www.foo.bar/index.html"
			*/

			end[-1] = '\0';
			urlurl  = strrchr(beg, '/') + 1;
			end[-1] = ':';
			break;
		}

		/* start at second slash */
		beg = ++end;

		/* end after last slash */
		while (*(++end) == '/') {;}

		if (urldup == NULL)
		{
			urldup	= g_strdup(urlstr);
			beg	= &urldup[beg - urlstr];
			end	= &urldup[end - urlstr];
			urlstr	= urldup;
		}

		/* move string (with '\0') down */
		memmove(beg, end, strlen(end) + 1);
	}

	beg	= urlstr;
	end	= &urlstr[colon];
	parent	= NULL;

	do
	{
		unsigned int	ch;
		gpointer	child;

		ch	= *end;
		*end	= '\0';

		child	= g_hash_table_lookup(folder_node, urlstr);

		if (child == NULL)
		{
			MainTreeAppendNode(&(notebook_list[kTreeList].store.tree), parent,
							beg, folder_icon[kIconFolder],
							folder_icon[kIconFolderOpen],
							child);

			g_hash_table_insert(folder_node, g_strdup(urlstr), child);
		}

		parent	= child;
		*end	= ch;
		beg 	= end;

		while (*(++beg) == '/') {;}

		if (beg == urlurl)
		{
			break;
		}

	}
	while ((end = strchr(beg, '/')) != NULL);

	if (*beg == '\0') 
	{
		beg--;
	}

	*dname += strlen(*dname) - strlen(beg);

	g_free(urldup);

	return parent;

} /* gpointer get_parent_node(char **dname, GHashTable *folder_node) */

static void nscrec_into_list_and_tree(nscache_record_t *rec, GHashTable *folder_node)
{
	gpointer	parent;
	gchar		*centry[10];
	gchar		size[15], mdtm[30], atm[30], exptm[30];

	const PixbufIcon *icon;

	if (gui_option[kOptionConvertToLower].active && rec->content_offset == 0)
	{
		g_strdown(rec->filename);
	}

	centry[kColumnURL]  = rec->urlstr;
	parent = get_parent_node(&centry[kColumnURL], folder_node);

	MainSetColumnEntryArray(centry, kColumnFile, rec->filename);
	MainSetColumnEntryArray(centry, kColumnType, rec->content_type);

	sprintf(size, "%d", rec->content_length);
	MainSetColumnEntryArray(centry, kColumnSize, size);

	MainSetColumnEntryArray(centry, kColumnEncoding, rec->content_encoding);
	MainSetColumnEntryArray(centry, kColumnCharset, rec->charset);

	MainSetColumnEntryArray(centry, kColumnModify,
		tl_time_to_str(mdtm, sizeof(mdtm), &(rec->last_modified)));

	MainSetColumnEntryArray(centry, kColumnAccess,
		tl_time_to_str(atm, sizeof(atm), &(rec->last_accessed)));

	MainSetColumnEntryArray(centry, kColumnExpire,
		tl_time_to_str(exptm, sizeof(exptm), &(rec->expires)));

	icon = guitl_get_icon_mime(rec->content_type);

	MainTreeInsertNode(&(notebook_list[kTreeList].store.tree),
					parent, centry, icon, rec);

	centry[kColumnURL] = rec->urlstr;

	MainListAppend(&(notebook_list[kSortedList].store.list),
						centry, icon, rec);

} /* static void nscrec_into_list_and_tree(nscache_record_t *rec, GHashTable *folder_node) */

static void gui_list_disable_entries(ListSelection *list, gpointer data)
{
#ifndef CONFIG_GTKCLIST_GTKCTREE
	g_free(list->name);
#endif
	list->name   = NULL;
	list->nscrec = NULL;
}

static void hash_entry_free(gpointer key, gpointer value, gpointer data)
{
	g_free(key);

#ifndef CONFIG_GTKCLIST_GTKCTREE
	g_free(value);
#endif
}

static void Open(GtkEntry *ew)
{
	gchar	*dbname, *name;
	GSList	*nscrec_list;

	dbname = NULL;

#ifdef HAVE_MSIE
	if (nscache_db.type != NSCACHE_RT_MSIE)
#endif
	{
		const gchar *fname = gtk_entry_get_text(ew);

		if (fname == NULL ||
			(dbname = utf8_to_filename(fname)) == NULL)
		{
			gui_err(gettext(invalid_filename_str));
			goto __done;
		}
	}

	/* results are invalid now */
	g_slist_foreach(search_dialog_list, (GFunc) search_disable_entries, NULL);

	/* free allocated memory */
	guitl_list_foreach_row_data(&(notebook_list[kSortedList].store.list),
						(GFunc) nscache_record_free, NULL);

	g_slist_foreach(list_selection_chain, (GFunc) gui_list_disable_entries, NULL);

#ifdef CONFIG_GTKCLIST_GTKCTREE
	g_slist_foreach(list_selection_chain,
		(GFunc) gui_clist_freeze_or_thaw, (gpointer) gtk_clist_freeze);
#endif
	GuiTreeClear(&(notebook_list[kTreeList].store.tree));
	GuiListClear(&(notebook_list[kSortedList].store.list));

#ifdef CONFIG_COLUMN_BUTTONS
	button_collexp.is_active = FALSE;
	gtk_widget_set_sensitive(button_collexp.collapse, FALSE);
	gtk_widget_set_sensitive(button_collexp.expand, FALSE);
#endif
	nscrec_list = nscache_read_cache(dbname);
	utf8_free(dbname);

	if(nscrec_list)
	{
		GHashTable *folder_node;

		folder_node = g_hash_table_new(g_str_hash, g_str_equal);

		g_slist_foreach(nscrec_list,
				(GFunc) nscrec_into_list_and_tree, folder_node);

		g_slist_free(nscrec_list);

		/* free strings */
		g_hash_table_foreach(folder_node, (GHFunc) hash_entry_free, NULL);
		g_hash_table_destroy(folder_node);

		GuiTreeColumnSort(&(notebook_list[kTreeList].store.tree),
					kColumnURL, gui_tree_compare_func, NULL);

		gui_list_sort(NULL, NULL);

		MainTreeExpandTopLevelNodes(&(notebook_list[kTreeList].store.tree));

		GuiTreeSelectFirstRow(&(notebook_list[kTreeList].store.tree));
		GuiListSelectFirstRow(&(notebook_list[kSortedList].store.list));
	}

	name = nscache_db.name;

	if (name)
	{
		gchar *path = (nscache_db.path).str;

		if (path == NULL)
		{
			path = name;
			name = NULL;
		}

		path = g_strconcat(path, name, NULL);
		dbname = utf8_from_filename(path, NULL);

		if (dbname)
		{
			gtk_entry_set_text(entry_db, dbname);
			utf8_free(dbname);
		}

		g_free(path); /* free here (if using GTK+-1.2) */
	}

#ifdef CONFIG_GTKCLIST_GTKCTREE
	g_slist_foreach(list_selection_chain,
		(GFunc) gui_clist_freeze_or_thaw, (gpointer) gtk_clist_thaw);
#endif
	/* records and size may have changed */
	update_entry_info(windata.toplist);

	print_node_info(NULL);

__done:
	gtk_option_menu_set_history(GTK_OPTION_MENU(nscache_type),
			(nscache_db.type > NSCACHE_RT_LAST ?
				NSCACHE_RT_FIRST : nscache_db.type));

} /* static void Open(GtkWidget *ew, const gchar *fname) */

#ifdef MULTI_DB_TYPES
static void TypeOpen(GtkWidget *w, gpointer index)
{
	nscache_db.type = (gint) index;

	Open(entry_db);
}
#endif

static void ToggleBool(GtkWidget *w, struct gui_option *opt)
{
	opt->active = GTK_CHECK_MENU_ITEM(w)->active;

	gprop_set_bool(opt->prop, opt->active);
}

static void SaveRC(GtkWidget *w, gpointer data)
{
	save_rc();
}

static void notebook_page_switched(GtkNotebook *nbook, GtkNotebookPage *page,
							gint nr, Windata *windat)
{
	ListSelection *list = &notebook_list[nr];

	GuiObjectSetUserData(windat->mainwin, list);

	windat->toplist = list;
	mainlist	= list;

	update_entry_info(list);
}

#ifdef CONFIG_GTKCLIST_GTKCTREE
void gui_list_row_selected(GtkCList* clist, gint row, gint col,
					GdkEventButton *e, ListSelection *list)
{
	if (row >= clist->rows)
	{
		gtk_clist_select_row(clist, clist->rows - 1, 0);
	}

	else if (gtk_clist_get_selectable(clist, row))
	{
		list->store.list.row = row;

		gtk_clist_get_pixtext(clist, row, kColumnURL,
					&(list->name), NULL, NULL, NULL);

		list->nscrec = gtk_clist_get_row_data(clist, row);

		if (windata.toplist == list)
		{
			update_entry_info(list);
		}
	}
}

void gui_list_row_unselected(GtkCList* clist, gint row, gint col,
					GdkEventButton *e, ListSelection *list)
{
	list->name   = NULL;
	list->nscrec = NULL;

	if(windata.toplist == list)
	{
		update_entry_info(list);
	}
}

void gui_tree_row_selected(GtkCTree* ctree, GList *node, gint col, ListSelection *list)
{
	list->store.tree.node = GTK_CTREE_NODE(node);

	gtk_ctree_node_get_pixtext(ctree, list->store.tree.node, kColumnURL,
						&(list->name), NULL, NULL, NULL);

	list->nscrec = gtk_ctree_node_get_row_data(ctree, list->store.tree.node);

#ifdef CONFIG_COLUMN_BUTTONS
	gui_collapse_expand_set_sensitive(list, &button_collexp);
#endif
	if(windata.toplist == list)
	{
		update_entry_info(list);
	}
}

void gui_tree_row_unselected(GtkCTree* ctree, GList *node, gint col, ListSelection *list)
{
	list->name   = NULL;
	list->nscrec = NULL;

#ifdef CONFIG_COLUMN_BUTTONS
	gtk_widget_set_sensitive(button_collexp.collapse, FALSE);
	gtk_widget_set_sensitive(button_collexp.expand, FALSE);
#endif
	if(windata.toplist == list)
	{
		update_entry_info(list);
	}
}

#else /* ! CONFIG_GTKCLIST_GTKCTREE */

void gui_tree_row_selected(GtkTreeSelection *selection, ListSelection *list)
{
	GtkTreeIter	iter;
	GtkTreeModel	*model;

	if (gtk_tree_selection_get_selected(selection, &model, &iter))
	{
		gint data_col;

		list->store.tree.iter = iter;

		g_free(list->name);

		data_col = list->type == kTreeList ?
				list->store.tree.data_column :
					list->store.list.data_column;

		gtk_tree_model_get(model, &iter, kColumnURL, &(list->name),
						data_col, &(list->nscrec), -1);
	}

	else
	{
		g_free(list->name);
		list->name   = NULL;
		list->nscrec = NULL;
	}

	if (windata.toplist == list)
	{
		update_entry_info(list);
	}
}

#endif /* #ifdef CONFIG_GTKCLIST_GTKCTREE .. #else */

#ifdef CONFIG_GTKCLIST_GTKCTREE

static void gui_tree_collapsed(GtkCTree *ctree, GtkCTreeNode *collapse_node, GuiTreeStore *tree)
{
	/*
		A folder was collapsed. If the selected
		entry is inside it, select the folder
		itself - avoids hidden entries.
	*/

	if (gtk_ctree_is_ancestor(ctree, collapse_node, tree->node))
	{
		gtk_ctree_select(ctree, collapse_node);
	}

#ifdef CONFIG_COLUMN_BUTTONS
	else if (button_collexp.is_active == FALSE)
	{
		gboolean can_collapse, can_expand;

		can_collapse = FALSE;
		can_expand   = FALSE;

		if (collapse_node == tree->node)
		{
			can_expand = TRUE;
		}

		else if (gtk_ctree_is_ancestor(ctree, tree->node, collapse_node))
		{
			can_collapse = TRUE;
			can_expand   = TRUE;
		}

		if (can_expand)
		{
			gtk_widget_set_sensitive(button_collexp.collapse, can_collapse);
			gtk_widget_set_sensitive(button_collexp.expand, can_expand);
		}
	}
#endif
}
	
#ifdef CONFIG_COLUMN_BUTTONS
static void gui_tree_expanded(GtkCTree *ctree, GtkCTreeNode *expand_node, GuiTreeStore *tree)
{
	if (button_collexp.is_active == FALSE)
	{
		if (expand_node == tree->node || gtk_ctree_is_ancestor(ctree, tree->node, expand_node))
		{
			gboolean can_expand = FALSE;

			guitl_tree_recursive(tree, TRUE,
				(GuiTreeStoreRecursiveFunc) gui_test_expand_recursive, &can_expand);

			gtk_widget_set_sensitive(button_collexp.expand, can_expand);
			gtk_widget_set_sensitive(button_collexp.collapse, TRUE);
		}
	}
}
#endif

#else /* ! CONFIG_GTKCLIST_GTKCTREE */

static void gui_tree_collapsed(GtkTreeView *view, GtkTreeIter *collapse_iter,
					GtkTreePath *collapse_path, GuiTreeStore *tree)
{
	if (gtk_tree_store_is_ancestor(tree->stree, collapse_iter, &(tree->iter)))
	{
		gtk_tree_selection_select_iter (tree->selection, collapse_iter);
	}
}
#endif /* #ifdef CONFIG_GTKCLIST_GTKCTREE .. #else */

#ifdef CONFIG_GTKCLIST_GTKCTREE

static void gui_list_resize_column(GtkWidget *w, gint col, gint width,
							gpointer data)
{
	int index, prop, type;

	index = column_tab[col]->index;
	type  = (int) data;
	prop  = type == kTreeList ?
			column_info[index].tree_width_prop :
				column_info[index].list_width_prop;

	gprop_set_int(prop, width);
}

#else /* ! CONFIG_GTKCLIST_GTKCTREE */

void gui_list_resize_column(GObject *column, GParamSpec *param, gpointer data)
{
	/*
		either main window or search dialog
		-----------------------------------
		kPropIntTreeColWidthUrl ... kPropIntTreeColWidthExp
		kPropIntListColWidthUrl ... kPropIntListColWidthExp
		kPropIntFindColWidthUrl ... kPropIntFindColWidthFile
	*/

	int prop, width;

	/* param->name -> "width" */
	g_object_get(column, param->name, &width, NULL);
	prop = (int) data;
	gprop_set_int(prop, width);
}

#endif /* #ifdef CONFIG_GTKCLIST_GTKCTREE .. #else */

static void gui_window_resize(GtkWidget *w, GtkAllocation *sallocation, gpointer data)
{
	/*
		either main window, search or info dialog
		-----------------------------------------
		kPropIntMainWinWidth + kPropIntMainWinHeight
		kPropIntFindWinWidth + kPropIntFindWinHeight
		kPropIntInfoWinWidth + kPropIntInfoWinHeight
	*/

	int prop = (int) data;

	gprop_set_int(prop, sallocation->width);
	gprop_set_int(prop + 1, sallocation->height);
}

static void adjust_recursive(GuiTreeStore *tree, nscache_record_t *rec, gboolean *sensitive)
{
	if (rec)
	{
		int cache_acc = file_access(rec);

		if (cache_acc > 0)
		{
			if (rec->content_offset == 0)
			{
				if (cache_acc & FILE_ACCESS_WRITE)
				{
					sensitive[kMenuReadOnly] = TRUE;
				}

				else
				{
					sensitive[kMenuReadWrite] = TRUE;
				}
#ifdef HAVE_LINK
				hardlink_enable = TRUE;
#endif
			}

			if (cache_acc & FILE_ACCESS_READ)
			{
				sensitive[kMenuSave] = TRUE;
			}
#ifndef CACHE_READONLY
			if (cache_acc & FILE_ACCESS_REMOVE)
			{
				sensitive[kMenuDelete] = TRUE;
			}
#endif
		}
	}

} /* static void adjust_recursive(tree, nscrec, sensitive) */

void AdjustMenu(GtkWidget *mni, ListSelection **selected)
{
	ListSelection		*list;
	nscache_record_t	*rec;
	gint			item;
	GtkWidget		**menu, **popup;
	gboolean		sensitive[kMenuItemsTotal];

	if (selected == NULL)
	{
		menu = &gui_menu_widget[kMenuItemsTotal];
		item = kMenuItemsTotal - 1;

		do
		{
			if (*(--menu) == mni)
			{
				break;
			}
		}
		while (--item >= 0);

		if (item < 0)
		{
			return;
		}

		list = windata.toplist;
	}

	else
	{
		list = *selected;
	}

#ifdef HAVE_LINK
	hardlink_enable = FALSE;
#endif
	memset(sensitive, '\0', sizeof(sensitive));

	rec = (list == NULL) ? NULL : list->nscrec;

	if (rec == NULL)
	{
		if (list && list->name && list->type == kTreeList)
		{
			if (GuiTreeGetLevel(&(list->store.tree)) > kTreeLevelService)
			{
#ifdef HAVE_MSIE
				/* no db path */
				if (nscache_db.type == NSCACHE_RT_MSIE ||
							access(nscache_db.path.str, F_OK) == 0)
#else
				/* still mounted ? */
				if (access(nscache_db.path.str, F_OK) == 0)
#endif
				{
					guitl_tree_recursive(&(list->store.tree), TRUE,
						(GuiTreeStoreRecursiveFunc) adjust_recursive,
											sensitive);
				}

			} /* if (GuiTreeGetLevel(list->store.tree) > kTreeLevelService) */

		} /* if (list && list->name && list->type == kTreeList) */

	} /* if (rec == NULL) */

	else
	{
		GSList *p;

		adjust_recursive(NULL, rec, sensitive);

		p = apassign_get_matching_viewers(rec->content_type);

		while (p)
		{
			apassign *e = (apassign *) p->data;

			if (e)
			{
				if (e->args_file != NULL && *e->args_file)
				{
					sensitive[kMenuViewFile] = sensitive[kMenuSave];
				}

				if (e->args_url != NULL && *e->args_url)
				{
					sensitive[kMenuViewURL] = TRUE;
				}
			}

			p = g_slist_remove(p, e);
		}

		sensitive[kMenuClipboard] = TRUE;

	} /* if (rec == NULL) .. else */

#ifndef CACHE_READONLY

	if (nscache_db.readonly || gui_option[kOptionReadOnly].active)
	{
		sensitive[kMenuDelete] = FALSE;
	}
#endif
	if (GuiListGetRows(&(notebook_list[kSortedList].store.list)) > 0)
	{
		sensitive[kMenuFind] = TRUE;
	}

	menu	= &gui_menu_widget[kMenuItemsTotal];
	popup	= &gui_popup_widget[kMenuItemsTotal];

	item = kMenuItemsTotal - 1;

	do
	{
		gtk_widget_set_sensitive(*(--menu),  sensitive[item]);
		gtk_widget_set_sensitive(*(--popup), sensitive[item]);

	}
	while (--item >= 0);

} /* static void AdjustMenu(ListSelection *list) */

gboolean gui_mouse_button_pressed(GtkWidget *widget, GdkEvent *event,
							GtkWidget *popup_menu)
{
	guint		button;
	guint32		time;
	GdkEventButton	*bevent;

	bevent = (GdkEventButton *) event;

	if (bevent == NULL)
	{
		time	= gdk_event_get_time((GdkEvent *) bevent);
		button	= 0;
		goto display_popup;
	}

	button = bevent->button;

	if (bevent->type == GDK_BUTTON_PRESS)
	{
		if (button == 3)
		{
			if ((windata.toplist)->name)
			{

#ifdef CONFIG_GTKCLIST_GTKCTREE
				gint     row, column;
				GtkCList *clist;

				clist = GTK_CLIST(widget);

				if (gtk_clist_get_selection_info(clist, (gint) bevent->x,
							(gint) bevent->y, &row, &column))
				{
					gtk_clist_select_row(clist, row, 0);
				}

#else /* ! CONFIG_GTKCLIST_GTKCTREE */

				GtkTreeView *view;
				GtkTreePath *path;

				view = GTK_TREE_VIEW(widget);

				if (gtk_tree_view_get_path_at_pos(view,
						(gint) bevent->x, (gint) bevent->y, &path,
						NULL, NULL, NULL))
				{
					GtkTreeSelection *selection;

					selection = gtk_tree_view_get_selection(view);
					gtk_tree_selection_select_path(selection, path);
					gtk_tree_path_free(path);
				}

#endif /* #ifdef CONFIG_GTKCLIST_GTKCTREE */

			} /* if ((windata.toplist)->name) */

			time = bevent->time;
display_popup:
			gtk_menu_popup(GTK_MENU(popup_menu), NULL, NULL, NULL, NULL, button, time);

			return TRUE;

		} /* if (button == 3) */

	} /* if (bevent->type == GDK_BUTTON_PRESS) */

	else if (bevent->type == GDK_2BUTTON_PRESS)
	{
		if (button == 1)
		{
			return ViewerDefault(VIEWER_FILE);
		}

		if (button == 2)
		{
			return ViewerDefault(VIEWER_URL);
		}
	}

	return FALSE;
}

#if GTK_MAJOR_VERSION == 1

gboolean gui_sf10_show_popup_menu(GtkWidget *widget, GdkEventKey *event,
							GtkWidget *popup_menu)
{
	if (event->state == GDK_SHIFT_MASK && event->keyval == GDK_F10)
	{
		return gui_mouse_button_pressed(widget, NULL, popup_menu);
	}

	return FALSE;
}

#else

gboolean gui_sf10_show_popup_menu(GtkWidget *widget, GtkWidget *popup_menu)
{
	return gui_mouse_button_pressed(widget, NULL, popup_menu);
}

#endif /* #if GTK_MAJOR_VERSION == 1 .. #else */


#ifdef CONFIG_SWAP_COLUMNS

static void gui_test_expand_recursive(GuiTreeStore *tree, nscache_record_t *rec, gboolean *can_expand)
{
	if (rec == NULL && *can_expand == FALSE)
	{
		gboolean is_expanded;

		GuiTreeGetExpanded(tree, is_expanded);

		if (is_expanded == FALSE)
		{
			*can_expand = TRUE;
		}

	} /* if (rec == NULL && *can_expand == FALSE) */
}

static void gui_collapse_expand_set_sensitive(ListSelection *list, CollapseExpandWidgets *cew)
{
		gboolean can_collapse, can_expand;

		can_collapse = can_expand = FALSE;

		if (list->name && list->nscrec == NULL)
		{
			GuiTreeGetExpanded(&(list->store.tree), can_collapse);

			if (can_collapse == FALSE)
			{
				can_expand = TRUE;
			}

			else
			{
				guitl_tree_recursive(&(list->store.tree), TRUE,
					(GuiTreeStoreRecursiveFunc) gui_test_expand_recursive, &can_expand);
			}
		}

		gtk_widget_set_sensitive(cew->collapse, can_collapse);
		gtk_widget_set_sensitive(cew->expand, can_expand);
}

static void gui_tree_collapse_recursive(GtkWidget *w, GuiTreeStore *tree)
{
#ifdef CONFIG_COLUMN_BUTTONS
	button_collexp.is_active = TRUE;
#endif
	GuiTreeCollapseRecursive(tree);

#ifdef CONFIG_COLUMN_BUTTONS
	button_collexp.is_active = FALSE;
	gtk_widget_set_sensitive(button_collexp.collapse, FALSE);
	gtk_widget_set_sensitive(button_collexp.expand, TRUE);
#endif
}

static void gui_tree_expand_recursive(GtkWidget *w, GuiTreeStore *tree)
{
#ifdef CONFIG_COLUMN_BUTTONS
	button_collexp.is_active = TRUE;
#endif
	GuiTreeExpandRecursive(tree);

#ifdef CONFIG_COLUMN_BUTTONS
	button_collexp.is_active = FALSE;
	gtk_widget_set_sensitive(button_collexp.collapse, TRUE);
	gtk_widget_set_sensitive(button_collexp.expand, FALSE);
#endif
}

static gboolean column_is_left_most(int position)
{
	while (--position > kColumnURL)
	{
		if (column_tab[position]->visible)
		{
			return FALSE;
		}
	}

	return TRUE;
}

static gboolean column_is_right_most(int position)
{
	while (++position <= kColumnInfoLastIndex)
	{
		if (column_tab[position]->visible)
		{
			return FALSE;
		}
	}

	return TRUE;
}

#ifdef CONFIG_COLUMN_BUTTONS

static void column_arrows_set_visibility(int nbook_page, int position)
{
	ColumnTab *col_tab = column_tab[position];
	ColumnArrows *arrow = &col_tab->arrow[nbook_page];

	if (column_is_left_most(position))
	{
		gtk_widget_hide(arrow->left);
	}
	else
	{
		gtk_widget_show(arrow->left);
	}

	if (column_is_right_most(position))
	{
		gtk_widget_hide(arrow->right);
	}
	else
	{
		gtk_widget_show(arrow->right);
	}
}

#endif /* CONFIG_COLUMN_BUTTONS */

static int column_get_right_visible(int position)
{
	while (++position <= kColumnInfoLastIndex)
	{
		if (column_tab[position]->visible)
		{
			return position;
		}
	}

	return -1;
}

static int column_get_left_visible(int position)
{
	while (--position >= kColumnURL)
	{
		if (column_tab[position]->visible)
		{
			break;
		}
	}

	return position;
}

#endif /* CONFIG_SWAP_COLUMNS */

static void column_set_visibility_list(gint nbook_page, gint index, gint active)
{
	int position;

	position = column_tab_array[index].position;

	if (nbook_page == kTreeList)
	{
		GuiTreeSetColumnVisiblility(&(notebook_list[kTreeList].store.tree),
								position, active);
	}

	else
	{
		GuiListSetColumnVisiblility(&(notebook_list[kSortedList].store.list),
								position, active);
	}

#ifdef CONFIG_COLUMN_BUTTONS

	if (index != kColumnURL)
	{
		int left_pos, right_pos;

		column_arrows_set_visibility(nbook_page, position);

		right_pos = column_get_right_visible(position);

		if (right_pos > 0)
		{
			column_arrows_set_visibility(nbook_page, right_pos);
		}

		left_pos = column_get_left_visible(position);

		if (left_pos > 0)
		{
			column_arrows_set_visibility(nbook_page, left_pos);
		}
	}

#endif /* CONFIG_COLUMN_BUTTONS */

} /* static void column_set_visibility_list(gint, gint, gint) */

static void column_menu_set_sensitive(gint active)
{
	static GtkWidget *last_visible_column = NULL;

	if (active == FALSE)
	{
		gint index, last;

		last	= -1;
		index	= kColumnInfoLastIndex;

		do
		{
			if (column_tab_array[index].visible)
			{
				if (last > kColumnURL)
				{
					return;
				}

				last = index;
			}
		}
		while (--index >= kColumnURL);

		last_visible_column = column_tab_array[last].menu;
	}

	if (last_visible_column)
	{
		gtk_widget_set_sensitive(last_visible_column, active);

		if (active)
		{
			last_visible_column = NULL;
		}
	}

} /* static void column_menu_set_sensitive(gint active) */

static void column_set_visibility(GtkWidget *menu_check_item, gpointer data)
{
	gint index  = (gint) data;
	gint active = GTK_CHECK_MENU_ITEM(menu_check_item)->active;

	gprop_set_bool(column_info[index].show_column_prop, active);

	column_tab_array[index].visible = active;
	column_menu_set_sensitive(active);

	column_set_visibility_list(kTreeList, index, active);
	column_set_visibility_list(kSortedList, index, active);

} /* static void column_set_visibility(GtkWidget*, gpointer) */

#ifdef CONFIG_SWAP_COLUMNS

#ifdef CONFIG_GTKCLIST_GTKCTREE

static void columns_do_swap(gint nbook_page, ColumnTab* col1, ColumnTab* col2)
{
	int      w1, w2;
	GtkCList *clist;

	if (nbook_page == kTreeList)
	{
		if (!gprop_get_int(column_info[col1->index].tree_width_prop, &w1))
		{
			w1 = column_info[col1->index].dwidth;
		}

		if (!gprop_get_int(column_info[col2->index].tree_width_prop, &w2))
		{
			w2 = column_info[col2->index].dwidth;
		}

		clist = GTK_CLIST(notebook_list[nbook_page].store.tree.ctree);
		gtk_clist_freeze(clist);
		guitl_ctree_swap_cols(notebook_list[nbook_page].store.tree.ctree,
							col1->position, col2->position);
	}

	else
	{
		if (!gprop_get_int(column_info[col1->index].list_width_prop, &w1))
		{
			w1 = column_info[col1->index].dwidth;
		}

		if (!gprop_get_int(column_info[col2->index].list_width_prop, &w2))
		{
			w2 = column_info[col2->index].dwidth;
		}

		clist = notebook_list[nbook_page].store.list.clist;
		gtk_clist_freeze(clist);
		guitl_clist_swap_cols(clist, col1->position, col2->position);
	}

	gtk_clist_set_column_justification(clist, col1->position,
					column_info[col1->index].justify);
	gtk_clist_set_column_justification(clist, col2->position,
					column_info[col2->index].justify);

	gtk_clist_set_column_width(clist, col1->position, w1);
	gtk_clist_set_column_width(clist, col2->position, w2);

	gtk_clist_thaw(clist);

#ifdef CONFIG_COLUMN_BUTTONS
	column_arrows_set_visibility(nbook_page, col1->position);
	column_arrows_set_visibility(nbook_page, col2->position);
#endif

} /* static void columns_do_swap(gint, ColumnTab*, ColumnTab*) */

#endif /* CONFIG_GTKCLIST_GTKCTREE */

static void columns_swap(ColumnTab* col1, gint direction)
{
	ColumnTab*	col2;
	gint		pos1, pos2;

	pos2 = (direction > 0) ?
			column_get_right_visible(col1->position) :
				column_get_left_visible(col1->position);
	if (pos2 < 0)
	{
		return;
	}

	/* swap pointer and position */

	col2 = column_tab[pos2];
	column_tab[pos2] = col1;

	pos1 = col1->position;
	column_tab[pos1] = col2;

	col1->position = col2->position;
	col2->position = pos1;

	/* swap columns */

#ifdef CONFIG_GTKCLIST_GTKCTREE
	columns_do_swap(kTreeList, col1, col2);
	columns_do_swap(kSortedList, col1, col2);
#else
	guitl_view_swap_cols(notebook_list[kTreeList].store.tree.view, pos1, pos2);
	guitl_view_swap_cols(notebook_list[kSortedList].store.list.view, pos1, pos2);
#endif
	gprop_set_int(column_info[col1->index].colnrprop, col1->position);
	gprop_set_int(column_info[col2->index].colnrprop, col2->position);

} /* static void columns_swap(ColumnTab* col1, gint direction) */

static void column_move_right(GtkWidget *widget, gpointer data)
{
	ColumnTab *column_tab;

#ifdef CONFIG_COLUMN_BUTTONS
	gtk_button_leave(GTK_BUTTON(widget));
	column_tab = data;
#else
	column_tab = *(ColumnTab **) data;
#endif
	columns_swap(column_tab, +1);
}

static void column_move_left(GtkWidget *widget, gpointer data)
{
	ColumnTab *column_tab;

#ifdef CONFIG_COLUMN_BUTTONS
	gtk_button_leave(GTK_BUTTON(widget));
	column_tab = data;
#else
	column_tab = *(ColumnTab **) data;
#endif
	columns_swap(column_tab, -1);
}

#ifdef CONFIG_COLUMN_BUTTONS

static GtkWidget *column_make_arrow(GtkArrowType direction, ColumnTab *col_tab)
{
	GtkWidget *arrow, *button;

	button = gtk_button_new();
	gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);

	GuiSignalConnect(button, "clicked",
		(direction == GTK_ARROW_LEFT ?
			column_move_left : column_move_right), col_tab);

	arrow = gtk_arrow_new(direction, GTK_SHADOW_OUT);
	gtk_container_add(GTK_CONTAINER(button), arrow);

	return button;

} /* static GtkWidget *column_make_arrow(GtkArrowType, ColumnTab *) */

static void column_make_widget(gint nbook_page, gint index)
{
	GtkWidget	*arrow, *label, *widget;
	GtkTable	*table;
	ColumnTab	*col_tab;

	col_tab = column_tab[index];

	widget = gtk_table_new(3, 1, FALSE);
	table = GTK_TABLE(widget);

	arrow = column_make_arrow(GTK_ARROW_LEFT, col_tab);
	col_tab->arrow[nbook_page].left = arrow;
	gtk_table_attach(table, arrow, 0, 1, 0, 1,
				GTK_FILL, GTK_FILL, 0, 0);

	label = guitl_label_new_gettext_utf8_from_str(column_info[col_tab->index].label);

	gtk_table_attach(table, label, 1, 2, 0, 1,
				GTK_EXPAND, GTK_EXPAND, 0, 0);

	arrow = column_make_arrow(GTK_ARROW_RIGHT, col_tab);
	col_tab->arrow[nbook_page].right = arrow;
	gtk_table_attach(table, arrow, 2, 3, 0, 1,
				GTK_FILL, GTK_FILL, 0, 0);

	if (nbook_page == kTreeList)
	{
		GuiTreeSetColumnWidget(&(notebook_list[kTreeList].store.tree),
									index, widget);
	}

	else
	{
		GuiListSetColumnWidget(&(notebook_list[kSortedList].store.list),
									index, widget);
	}

	gtk_widget_show_all(widget);

	column_arrows_set_visibility(nbook_page, index);

} /* static void column_make_widget(gint nbook_page, gint index) */

#else /* ! CONFIG_COLUMN_BUTTONS */

#ifdef CONFIG_COLUMN_POPUP

static gboolean gui_popup_collapse_expand(GtkWidget *w, GdkEvent *event, GtkWidget **popup_menu)
{
	GdkEventButton *bevent = (GdkEventButton *) event;

	if (bevent->type == GDK_BUTTON_PRESS && bevent->button == 3)
	{
		gui_collapse_expand_set_sensitive(&notebook_list[kTreeList], &popup_collexp);

		gtk_menu_popup(GTK_MENU(*popup_menu), NULL, NULL, NULL, NULL,
							bevent->button, bevent->time);
	}

	return TRUE;
}

static gboolean gui_popup_colswap(GtkWidget *widget, GdkEvent *event, ColumnTab *col_tab)
{
	GdkEventButton *bevent = (GdkEventButton *) event;

	if (bevent->type == GDK_BUTTON_PRESS && bevent->button == 3)
	{
		gboolean sensitive;

#ifdef CONFIG_GTKCLIST_GTKCTREE
		col_tab = *(ColumnTab **) col_tab;
#endif
		popup_colswap.selected = col_tab;

		sensitive = !column_is_left_most(col_tab->position);
		gtk_widget_set_sensitive(popup_colswap.left, sensitive);

		sensitive = !column_is_right_most(col_tab->position);
		gtk_widget_set_sensitive(popup_colswap.right, sensitive);

		gtk_menu_popup(GTK_MENU(popup_colswap.menu), NULL, NULL, NULL, NULL,
							bevent->button, bevent->time);
	}

	return TRUE;
}

static void column_make_widget(gint nbook_page, gint index)
{
	gint		position;
	GtkWidget	*label, *button;

	position = column_tab[index]->index;
	label = guitl_label_new_gettext_utf8_from_str(column_info[position].label);

	if (nbook_page == kTreeList)
	{
		GuiTreeSetColumnWidget(&(notebook_list[kTreeList].store.tree),
								index, label);
		GuiTreeColumnTitleActive(&(notebook_list[kTreeList].store.tree),
								index, TRUE);
	}

	else
	{
		GuiListSetColumnWidget(&(notebook_list[kSortedList].store.list),
								index, label);
		GuiListColumnTitleActive(&(notebook_list[kSortedList].store.list),
								index, TRUE);
	}

	button = gtk_widget_get_ancestor(label, GTK_TYPE_BUTTON);

	if (button)
	{
		if (index == kColumnURL)
		{
			GuiSignalConnect(button, "button-press-event",
				gui_popup_collapse_expand, &popup_collexp.menu);
		}

		else
		{
			gpointer data;

#ifdef CONFIG_GTKCLIST_GTKCTREE
			/* widget and button may differ */
			data = &column_tab[index];
#else
			/* widget and button always equal */
			data = column_tab[index];
#endif
			GuiSignalConnect(button, "button-press-event",
							gui_popup_colswap, data);
		}
	}

	gtk_widget_show(label);
}

#endif /* CONFIG_COLUMN_POPUP */

#endif /* #ifdef CONFIG_COLUMN_BUTTONS .. #else */

#else /* ! CONFIG_SWAP_COLUMNS */

#define column_make_widget(nbook_page, index) \
		MainSetColumnTitle(nbook_page, index)

#endif /* #if CONFIG_SWAP_COLUMNS .. #else */


static GtkWidget *gui_build_notebook_page(GtkWidget *notebook, GtkWidget *parent)
{
	static const gchar *page_label[] =
	{
		gettext_nop("Tree view"),
		gettext_nop("Sorted view")
	};

	gint		type;
	GtkWidget	*label, *swin;

	swin = gtk_scrolled_window_new(NULL, NULL);
	GuiWidgetSetSizeRequest(swin, -1, 150);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(swin),
					GTK_POLICY_ALWAYS, GTK_POLICY_ALWAYS);
	if (parent == NULL)
	{
		parent = swin;
		type = kTreeList;
	}

	else
	{
		gtk_box_pack_start(GTK_BOX(parent), swin, TRUE, TRUE, 1);
		type = kSortedList;
	}

	label = guitl_label_new_gettext_utf8_from_str(page_label[type]);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), parent, label);

	list_selection_chain =
		g_slist_append(list_selection_chain, &notebook_list[type]);

	notebook_list[type].type = type;

	return swin;
}

static void gui_build_sorted_list(GtkWidget *notebook)
{
	static const gchar *stype_label[] =
	{
		gettext_nop("URL"),		/* STYPE_URL  */
		gettext_nop("Local file"),	/* STYPE_FILE  */
		gettext_nop("Size"), 		/* STYPE_SIZE */
		gettext_nop("Access time"),	/* STYPE_ATM  */
		gettext_nop("Mime type")	/* STYPE_MIME */
	};

	static gint stype_array[STYPE_MIME + 1] = { 0 };

	GuiListStore	*list;
	GtkBox		*hbox;
	GtkWidget	*label, *mi, *menu, *swin, *stype_menu, *vbox, *widget;
	gchar		*str;
	gint		i, sort_type;

	vbox = gtk_vbox_new(FALSE, 0);

	widget = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_start(GTK_BOX(vbox), widget, FALSE, FALSE, 1);

	hbox = GTK_BOX(widget);
	label = guitl_label_new_gettext_utf8_from_str(gettext_nop("Sort by: "));
	gtk_box_pack_start(hbox, label, FALSE, FALSE, 1);
	
	stype_menu = gtk_option_menu_new();
	gtk_box_pack_start(hbox, stype_menu, FALSE, FALSE, 1);

	menu = gtk_menu_new();

	i = STYPE_URL;

	do
	{
		str = guitl_gettext_utf8_from_str(stype_label[i]);
		mi = gtk_menu_item_new_with_label(str);
		utf8_free(str);

		GuiMenuAppend(menu, mi);
		gtk_widget_show(mi);

		stype_array[i] = i;
		GuiSignalConnect(mi, "activate", gui_list_sort, &stype_array[i]);
	}
	while (++i <= STYPE_MIME);

	gtk_option_menu_set_menu(GTK_OPTION_MENU(stype_menu), menu);

	if (gprop_get_int(kPropIntSortType, &sort_type) == FALSE)
	{
		i = sort_type = STYPE_URL;
	}

	else
	{
		i = sort_type & ~STYPE_REVERS_MASK;

		if (i < STYPE_URL || i > STYPE_MIME)
		{
			i = sort_type = STYPE_URL;
		}
	}

	stype_array[i] = sort_type;
	nscache_sort_type = sort_type;
	gtk_option_menu_set_history(GTK_OPTION_MENU(stype_menu), i);

	swin = gui_build_notebook_page(notebook, vbox);
	list = &(notebook_list[kSortedList].store.list);

	MainListNew(list);
	GuiListContainerAdd(list, swin);

	i = kColumnURL;

	do
	{
		int w;

		if (i == kColumnURL)
		{

#ifndef CONFIG_GTKCLIST_GTKCTREE
			GuiListColumnPixtextNew(list, kColumnURL, kColumnPixMime);
#endif
			MainSetColumnTitle(kSortedList, i);
		}

		else
		{

#ifndef CONFIG_GTKCLIST_GTKCTREE
			GuiListColumnTextNew(list, i,
				column_info[column_tab[i]->index].justify);
#endif
			column_make_widget(kSortedList, i);
		}

#ifdef CONFIG_GTKCLIST_GTKCTREE
		gtk_clist_set_column_justification(list->clist, i,
					column_info[column_tab[i]->index].justify);
#else
		GuiListSignalConnectResizeColumn(list, i, gui_list_resize_column,
			(gpointer) column_info[column_tab[i]->index].list_width_prop);

		GuiListSetColumnAlignment(list, i, 0.5);
#endif
		GuiListSetColumnAutoResize(list, i, FALSE);
		GuiListSetColumnResizable(list, i, TRUE);

		GuiListSetColumnWidth(list, i,
			gprop_get_int(column_info[column_tab[i]->index].list_width_prop, &w) ?
								w : column_info[i].dwidth);
	}
	while (++i <= kColumnInfoLastIndex);

	GuiListSetSelectionMode(list, GTK_SELECTION_BROWSE);
	GuiListColumnTitlesShow(list);

#ifdef CONFIG_COLUMN_POPUP
	GuiListColumnTitleActive(list, kColumnURL, FALSE);
#else
	GuiListColumnTitlesPassive(list);
#endif

#ifdef CONFIG_GTKCLIST_GTKCTREE
	GuiListSignalConnectResizeColumn(list,
				gui_list_resize_column, (gpointer) kSortedList);
#endif
	GuiListSignalConnectSelectedRow(list,
				gui_list_row_selected, &notebook_list[kSortedList]);
	GuiListSignalConnectUnselectedRow(list,
				gui_list_row_unselected, &notebook_list[kSortedList]);

} /* static void gui_build_sorted_list(GtkWidget *notebook) */

#ifdef CONFIG_SWAP_COLUMNS

#ifdef CONFIG_COLUMN_BUTTONS

static GtkWidget *tree_list_make_expand_collapse_button(gint pixid)
{
	GtkWidget *button;

	button = guitl_get_icon_button(pixid);
	gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);

	GuiSignalConnect(button, "clicked",
		(gpointer) (pixid == kIconButtonExpand ?
			gui_tree_expand_recursive : gui_tree_collapse_recursive),
				&(notebook_list[kTreeList].store.tree));
	return button;

} /* static GtkWidget *tree_list_make_expand_collapse_button(gint type) */

static void tree_list_make_first_column(GuiTreeStore *tree)
{
	GtkWidget *button, *label, *widget;
	GtkTable  *table;

	widget = gtk_table_new(3, 1, FALSE);
	table  = GTK_TABLE(widget);

	button = tree_list_make_expand_collapse_button(kIconButtonExpand);

#ifdef CONFIG_COLUMN_BUTTONS
	button_collexp.is_active = FALSE;
	button_collexp.expand    = button;
#endif
	gtk_table_attach(table, button, 0, 1, 0, 1,
				GTK_FILL, GTK_FILL, 0, 0);
 
	button = tree_list_make_expand_collapse_button(kIconButtonCollapse);

#ifdef CONFIG_COLUMN_BUTTONS
	button_collexp.collapse = button;
#endif
	gtk_table_attach(table, button, 1, 2, 0, 1,
				GTK_FILL, GTK_FILL, 0, 0);

	label = guitl_label_new_gettext_utf8_from_str(column_info[kColumnURL].label);
	gtk_table_attach(table, label, 2, 3, 0, 1,
				GTK_FILL, GTK_FILL, 1, 1);

	GuiTreeSetColumnWidget(tree, kColumnURL, widget);
	gtk_widget_show_all(widget);

} /* static void tree_list_make_first_column(GuiTreeStore *tree) */

#else /* ! CONFIG_COLUMN_BUTTONS */

#define tree_list_make_first_column(tree) \
		column_make_widget(kTreeList, kColumnURL)

#endif /* CONFIG_COLUMN_BUTTONS */

#else /* ! CONFIG_SWAP_COLUMNS */

#define tree_list_make_first_column(tree) \
		MainSetColumnTitle(kTreeList, kColumnURL)

#endif /* #ifdef CONFIG_SWAP_COLUMNS .. #else */

static void gui_build_tree_list(GtkWidget *notebook)
{
	GuiTreeStore	*tree;
	GtkWidget	*swin;
	gint		i;

	swin = gui_build_notebook_page(notebook, NULL);
	tree = &notebook_list[kTreeList].store.tree;

	MainTreeNew(tree);
	GuiTreeContainerAdd(tree, swin);

	i = kColumnURL;

	do
	{
		int w;

		if (i == kColumnURL)
		{
#ifndef CONFIG_GTKCLIST_GTKCTREE
			GuiTreeColumnPixtextNew(tree, kColumnURL, kColumnPixMime,
							kColumnPixClose, kColumnPixOpen);
#endif
			tree_list_make_first_column(tree);
		}

		else
		{
#ifndef CONFIG_GTKCLIST_GTKCTREE
			GuiTreeColumnTextNew(tree, i,
				column_info[column_tab[i]->index].justify);
#endif
			column_make_widget(kTreeList, i);
		}

#ifdef CONFIG_GTKCLIST_GTKCTREE
		gtk_clist_set_column_justification(GTK_CLIST(tree->ctree), i,
					column_info[column_tab[i]->index].justify);
#else
		GuiTreeSignalConnectResizeColumn(tree, i, gui_list_resize_column,
			(gpointer) column_info[column_tab[i]->index].tree_width_prop);

		GuiTreeSetColumnAlignment(tree, i, 0.5);
#endif
		GuiTreeSetColumnAutoResize(tree, i, FALSE);
		GuiTreeSetColumnResizable(tree, i, TRUE);

		GuiTreeSetColumnWidth(tree, i,
			gprop_get_int(column_info[column_tab[i]->index].tree_width_prop, &w) ?
								w : column_info[i].dwidth);
	}
	while (++i <= kColumnInfoLastIndex);

	GuiTreeSetSelectionMode(tree, GTK_SELECTION_BROWSE);
	GuiTreeColumnTitlesShow(tree);

#ifndef CONFIG_COLUMN_POPUP
	GuiTreeColumnTitlesPassive(tree);
#endif

#ifdef CONFIG_GTKCLIST_GTKCTREE
	gtk_ctree_set_expander_style(tree->ctree, GTK_CTREE_EXPANDER_SQUARE);
	gtk_ctree_set_line_style(tree->ctree, GTK_CTREE_LINES_DOTTED);

	GuiTreeSignalConnectResizeColumn(tree,
				gui_list_resize_column, (gpointer) kTreeList);
#endif
	GuiTreeSignalConnectSelectedRow(tree,
				gui_tree_row_selected, &notebook_list[kTreeList]);
	GuiTreeSignalConnectUnselectedRow(tree,
				gui_tree_row_unselected, &notebook_list[kTreeList]);

#ifdef CONFIG_COLUMN_BUTTONS
	GuiTreeSignalConnectAfterCollapse(tree, gui_tree_collapsed, tree);
	GuiTreeSignalConnectAfterExpand(tree, gui_tree_expanded, tree);
#else
	GuiTreeSignalConnectCollapse(tree, gui_tree_collapsed, tree);
#endif

} /* static void gui_build_tree_list(GtkWidget *notebook) */

#ifdef HAVE_GLOB

static gint gui_completition_destroy(GtkWidget *widget, GdkEventAny *e, GdkWindow *window)
{
	GdkWindow *wnd;

	if (e->type == GDK_KEY_PRESS && ((GdkEventKey *) e)->keyval == GDK_Tab)
	{
		return TRUE;
	}

	gtk_widget_destroy(widget);

	e = (GdkEventAny *) gdk_event_copy((GdkEvent *) e);

	if (e->type == GDK_BUTTON_PRESS)
	{
		gint x, y;
		window = gdk_window_at_pointer(&x, &y);
		((GdkEventButton *) e)->x = x;
		((GdkEventButton *) e)->y = y;
	}

	wnd = e->window;
	e->window = window;
	gdk_event_put((GdkEvent *) e);
	e->window = wnd;

	gdk_event_free((GdkEvent *) e);

	return TRUE;

} /* static gint gui_completition_destroy(GtkWidget*, GdkEventAny*, GdkWindow*) */

static gint gui_completition_expose(GtkWidget *label, GdkEvent *e, GtkWidget *widget)
{
	gdk_draw_rectangle(widget->window,
		(widget->style)->fg_gc[GTK_STATE_NORMAL],
		FALSE,
		(widget->allocation).x,
		(widget->allocation).y,
		(widget->allocation).width - 1,
		(widget->allocation).height - 1);

	return TRUE;

} /* static gint gui_completition_expose(GtkWidget*, GdkEvent*, GtkWidget*) */

static gint gui_completition_show_matches(GtkWidget *entry, gchar **fnamev)
{
	static const char spaces[] = "   ";
	static GtkStyle *completion_style = NULL;

	char		*str;
	gint		x, y;
	GtkWidget	*label, *widget;
	GtkStyle 	*style;
	StringBuf	strbuf;

	stringbuf_new(&strbuf, 0);

	while ((str = *(fnamev++)) != NULL)
	{
		int	slen;
		char	*tmp;

		tmp = utf8_from_filename(str, &slen);

		if (tmp)
		{
			stringbuf_append(&strbuf, tmp, slen);
			utf8_free(tmp);
			stringbuf_append(&strbuf, spaces, sizeof(spaces) - 1);
		}
	}

	if (strbuf.len == 0)
	{
		stringbuf_delete(&strbuf);
		return -1;
	}

	stringbuf_truncate(&strbuf, strbuf.len - (sizeof(spaces) - 1));

	label = gtk_label_new(strbuf.str);
	stringbuf_delete(&strbuf);

	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
	gtk_misc_set_alignment(GTK_MISC(label), 0.5, 0.5);

	widget  = gtk_window_new(GTK_WINDOW_POPUP);
	gtk_container_add(GTK_CONTAINER(widget), label);
	gtk_container_set_border_width(GTK_CONTAINER(widget), 4);
	gtk_widget_set_events(widget, GDK_BUTTON_PRESS_MASK | GDK_KEY_PRESS_MASK);

	style = completion_style;

	if (style == NULL)
	{
		style = gtk_style_copy(gtk_widget_get_style(label));
		gdk_color_parse("Black", &style->fg[GTK_STATE_NORMAL]);
		gdk_color_parse("LightYellow", &style->bg[GTK_STATE_NORMAL]);
		completion_style = style;
	}

	gtk_widget_set_style(label, style);
	gtk_widget_set_style(widget, style);

	GuiSignalConnectAfter(label, "expose-event",
					gui_completition_expose, widget);
	GuiSignalConnect(widget, "button-press-event",
					gui_completition_destroy, entry->window);
	GuiSignalConnect(widget, "key-press-event",
					gui_completition_destroy, entry->window);
	gtk_widget_realize(widget);
	gdk_window_get_origin (entry->window, &x, &y);
	GuiWindowMove(GTK_WINDOW(widget), x, y + entry->allocation.height);

	gtk_widget_show_all(widget);
	
	if (gdk_pointer_grab(widget->window, FALSE, GDK_BUTTON_PRESS_MASK,
						NULL, NULL, GDK_CURRENT_TIME) ||
			gdk_keyboard_grab(widget->window, FALSE, GDK_CURRENT_TIME))
	{
		gtk_widget_destroy(widget);
		gdk_beep();
	}

	return 0;

} /* static gint gui_completition_show_matches(GtkWidget *entry, gchar **fnamev) */

static gboolean gui_filename_completition(GtkWidget *ew, GdkEventKey *event, gpointer data)
{
	gint		chrpos;
	gchar		*dbname, *fname;
	const gchar	*errstr;
	GtkEditable	*editable;

	if (event->keyval != GDK_Tab)
	{
		return FALSE;
	}

	GuiSignalEmitStopByName(ew, "key-press-event");

	editable = GTK_EDITABLE(ew);

	if (GuiEditableGetSelectionStart(editable, &chrpos))
	{
		gtk_editable_delete_selection(editable);
	}

	else
	{
		chrpos = gtk_editable_get_position(editable);
	}

	fname  = gtk_editable_get_chars(editable, 0, chrpos);
	dbname = utf8_to_filename(fname);
	errstr = gettext(invalid_filename_str);

	if (dbname)
	{
		CompletitonData	*ptr;

		ptr = tl_fname_completion(dbname);
		utf8_free(dbname);

		if (ptr == NULL)
		{
			errstr = (errno != 0) ?
					strerror(errno) :
						gettext("No match found");
		}

		else
		{
			gint	slen;
			gchar	*str;

			str = utf8_from_filename(ptr->append, &slen);

			if (str)
			{
				gint		chrlen;
				const gchar	*tmperr;

				chrlen = utf8_strlen(str);
				tmperr = NULL;

				if (chrlen <= 0)
				{
					if (ptr->fnamec == 1)
						tmperr = gettext("Match");
				}

				else
				{
					gint ipos;

					if (ptr->fnamec == 1)
					{
						if (*utf8_offset_to_pointer(str, chrlen - 1) == '/')
						{
							gchar *s = gtk_editable_get_chars(editable,
									chrpos, chrpos + 1);
							if (*s == '/')
							{
								slen--;
							}

							g_free(s);
						}
					}

					ipos = chrpos;
					gtk_editable_insert_text(editable, str, slen, &ipos);
					gtk_editable_set_position(editable, chrlen + chrpos);

				} /* if (chrlen <= 0) .. else */

				utf8_free(str);

				if (ptr->fnamec > 1 &&
					gui_completition_show_matches(ew, ptr->fnamev))
				{
					tmperr = errstr;
					
				}

				errstr = tmperr;

			} /* if (str) */

			tl_completion_free(ptr);

		} /* if (ptr == NULL) .. else */

	} /* if (dbname) */

	g_free(fname); /* free here (if using GTK+-1.2) */

	gui_err(errstr);

	return TRUE;

} /* static gboolean gui_filename_completition(GtkWidget *, GdkEventKey *, gpointer) */

#endif /* HAVE_GLOB */

static GtkWidget *gui_build_entry_select_browse(void)
{
	GtkBox		*box;
	GtkWidget	*button, *ew, *frame, *label, *menu, *widget;
	gint		i;

	widget = gtk_hbox_new(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(widget), 4);

	frame = guitl_frame_new_add_child(NULL, widget);
	gtk_container_set_border_width(GTK_CONTAINER(frame), 2);

	box = GTK_BOX(widget);
	label = guitl_label_new_gettext_utf8_from_str(gettext_nop("Cache index file: "));
	gtk_box_pack_start(box, label, FALSE, FALSE, 5);

	ew	 = gtk_entry_new();
	entry_db = GTK_ENTRY(ew);
	gtk_box_pack_start(box, ew, TRUE, TRUE, 1);

	GuiSignalConnect(ew, "activate", Open, NULL);

#ifdef HAVE_GLOB
	GuiSignalConnect(ew, "key-press-event", gui_filename_completition, NULL);
#endif
	nscache_type = gtk_option_menu_new();
	gtk_box_pack_start(box, nscache_type, FALSE, FALSE, 2);

#ifndef MULTI_DB_TYPES
	gtk_widget_set_sensitive(nscache_type, FALSE);
#endif
	menu	= gtk_menu_new();
	i	= NSCACHE_RT_FIRST;

	do
	{
		GtkWidget *mi;

		mi = guitl_get_icon_menu(i);
		GuiMenuAppend(menu, mi);

#ifdef MULTI_DB_TYPES
		GuiSignalConnect(mi, "activate", TypeOpen, (gpointer) i);
#endif
	}
	while (++i <= NSCACHE_RT_LAST);

	gtk_option_menu_set_menu(GTK_OPTION_MENU(nscache_type), menu);
	gtk_option_menu_set_history(GTK_OPTION_MENU(nscache_type),
							nscache_db.type);
	button = guitl_get_icon_button(kIconButtonOpen);
	gtk_box_pack_start(box, button, FALSE, FALSE, 1);
	GuiSignalConnect(button, "clicked", Browse, NULL);

	return frame;

} /* static GtkWidget *gui_build_entry_select_browse(void) */

static void gui_build_submenu_checkitems(GtkWidget *submenu)
{
	int index = kColumnURL;

	do
	{
		GtkWidget	*mi;
		gchar		*str;

		str = guitl_gettext_utf8_from_str(column_info[index].label);
		mi = gtk_check_menu_item_new_with_label(str);
		utf8_free(str);

		GuiMenuAppend(submenu, mi);
		column_tab_array[index].menu = mi;
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi),
						column_tab_array[index].visible);
		gaccel_bind_widget(column_info[index].accel_id,
						column_info[index].label, mi, NULL);

		column_set_visibility(mi, (gpointer) index);

		GuiSignalConnect(mi, "activate", column_set_visibility, (gpointer) index);
	}
	while (++index <= kColumnInfoLastIndex);

	column_menu_set_sensitive(FALSE);

} /* static void gui_build_submenu_checkitems(submenu, accel_group) */

static void gui_build_menu_item(const GuiMenuItem *item, GtkWidget *menu, GtkWidget *window)
{
	for (;;)
	{
		GtkWidget	*mi, *submenu;
		gchar		*str;

		str = guitl_gettext_utf8_from_str(item->label);

		switch (item->type)
		{
			case kMTypeItem:
			case kMTypeSubmenu:
			case kMTypeSubmenuCheckItem:
				mi = gtk_menu_item_new_with_label(str);

				if (item->type == kMTypeItem)
				{
					break;
				}

				submenu = gtk_menu_new();
				gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), submenu);
				gtk_widget_realize(submenu);

				if (item->callback)
				{
					GuiSignalConnect(submenu, "show",
						item->callback, item->data);
				}

				if (item->type == kMTypeSubmenuCheckItem)
				{
					gui_build_submenu_checkitems(submenu);
				}

				break;

			case kMTypeCheckItem:
				mi = gtk_check_menu_item_new_with_label(str);
				gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi),
						((struct gui_option *) item->data)->active);
				break;

			case kMTypeSeparator:
				mi = gtk_menu_item_new();
				break;

			default:
				return;
		}

		GuiMenuAppend(menu, mi);
		gtk_widget_show(mi);

		utf8_free(str);

		if (item->accel_id != kGAccelKeyNone)
		{
			gaccel_bind_widget(item->accel_id, item->label, mi, window);
		}

		if (item->callback && item->type != kMTypeSubmenu)
		{
			GuiSignalConnect(mi, "activate", item->callback, item->data);
		}

		if (item->storage)
		{
			*item->storage = mi;
		}

		item++;

	} /* for (;;) */

} /* static void gui_build_menu_item(const GuiMenuItem*, GtkWidget*, GtkWidget*) */

static GtkWidget *gui_build_menu(GtkWidget *window)
{
	static const struct
	{
		const gchar		*label;
		const GuiMenuItem	*entries;
		gboolean		adjust_menu;

	} gui_menu_array[] = {

		{gettext_nop("File"), menuFile, FALSE},
		{gettext_nop("View"), menuView, TRUE},
		{gettext_nop("Edit"), menuEdit, TRUE},
		{gettext_nop("Options"), menuOptions, FALSE},
		{gettext_nop("Help"), menuHelp, FALSE},
		{NULL}

	}, *gui_menu;

	gchar		*str;
	GtkWidget	*mbar, *mbb, *menu;

	gprop_get_bool(gui_option[kOptionConvertToLower].prop,
				&gui_option[kOptionConvertToLower].active);
#ifndef CACHE_READONLY
	gprop_get_bool(gui_option[kOptionReadOnly].prop,
				&gui_option[kOptionReadOnly].active);
#endif
	gprop_get_bool(gui_option[kOptionSaveConfigOnExit].prop,
				&gui_option[kOptionSaveConfigOnExit].active);

	mbar = gtk_menu_bar_new();
	GuiMenuBarDisableShadow(mbar);

	gui_menu = gui_menu_array;

	do
	{
		str = guitl_gettext_utf8_from_str(gui_menu->label);
		mbb = gtk_menu_item_new_with_label(str);
		utf8_free(str);
		GuiMenuBarAppend(mbar, mbb);

		if (gui_menu->adjust_menu)
		{
			GuiSignalConnect(mbb, "activate",
					AdjustMenu, &(windata.toplist));
		}

		menu = gtk_menu_new();
		gtk_widget_realize(menu);

		gtk_menu_item_set_submenu(GTK_MENU_ITEM(mbb), menu);

		gui_build_menu_item(gui_menu->entries, menu, window);

		gui_menu++;
	}
	while (gui_menu->label);

	GuiMenuItemSetRightJustified(GTK_MENU_ITEM(mbb), TRUE);

	menu		= gtk_menu_new();
	gui_popup_menu	= menu;

	GuiSignalConnect(menu, "show", AdjustMenu, &(windata.toplist));

	gtk_widget_realize(menu);

	gui_build_menu_item(menuPopup, menu, NULL);

#ifdef CONFIG_COLUMN_POPUP

	menu = gtk_menu_new();
	popup_colswap.menu = menu;

	str = guitl_gettext_utf8_from_str(gettext_nop("left"));
	popup_colswap.left = gtk_menu_item_new_with_label(str);
	utf8_free(str);

	GuiMenuAppend(menu, popup_colswap.left);
	GuiSignalConnect(popup_colswap.left, "activate",
				column_move_left, &(popup_colswap.selected));

	str = guitl_gettext_utf8_from_str(gettext_nop("right"));
	popup_colswap.right = gtk_menu_item_new_with_label(str);
	utf8_free(str);

	GuiMenuAppend(menu, popup_colswap.right);
	GuiSignalConnect(popup_colswap.right, "activate",
				column_move_right, &(popup_colswap.selected));

	gtk_widget_show_all(menu);

	menu = gtk_menu_new();
	popup_collexp.menu = menu;

	str = guitl_gettext_utf8_from_str(gettext_nop("collapse"));
	mbb = gtk_menu_item_new_with_label(str);
	utf8_free(str);

	popup_collexp.collapse = mbb;
	GuiMenuAppend(menu, mbb);
	GuiSignalConnect(mbb, "activate", gui_tree_collapse_recursive,
				&(notebook_list[kTreeList].store.tree));

	str = guitl_gettext_utf8_from_str(gettext_nop("expand"));
	mbb = gtk_menu_item_new_with_label(str);
	utf8_free(str);

	popup_collexp.expand = mbb;
	GuiMenuAppend(menu, mbb);
	GuiSignalConnect(mbb, "activate", gui_tree_expand_recursive,
				&(notebook_list[kTreeList].store.tree));

	gtk_widget_show_all(menu);

#endif /* CONFIG_COLUMN_POPUP */

	return mbar;

} /* static GtkWidget *gui_build_menu(GtkWidget *window) */

static gint gui_error_msg_expose(GtkWidget *label, GdkEventExpose *e, gpointer p)
{
	gdk_draw_rectangle(label->window,
		(label->style)->bg_gc[GTK_STATE_NORMAL],
		TRUE,
		(e->area).x,
		(e->area).y,
		(e->area).width,
		(e->area).height);

	return FALSE;

} /* static gint gui_error_msg_expose(GtkWidget*, GdkEvent*, gpointer) */

static GtkWidget *gui_build_msg_tmr(GtkWidget *box, struct msg_tmr *msg)
{
	GtkWidget *frame, *label;

	label = gtk_label_new(NULL);
	msg->label = GTK_LABEL(label);
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);

	frame = guitl_frame_new_add_child(NULL, label);

	msg->tmr_id = 0;
	msg->hdl_id = 0;
	msg->frame  = frame;

	gtk_container_add(GTK_CONTAINER(box), frame);

	return label;
}

static void _noop()
{
	return;
}

static void gui_init(char *dbname)
{
	GtkBox		*vbox;
	GtkWidget	*entry, *label, *menu_bar, *notebook, *widget, *window;
	GtkStyle	*style;

	g_log_set_handler("Gdk", G_LOG_LEVEL_MASK, (GLogFunc) _noop, NULL);

#ifdef FOCUS_IN_BUGFIX
	windata._net_wm_state = gdk_atom_intern( _NET_WM_STATE, TRUE );
	windata.ismapped = NULL;
	windata.unmapped = NULL;
#endif

	windata.topwin = NULL;

	window = gui_window_new("MozCache", kPropIntMainWinWidth);
	windata.mainwin = window;

	GuiSignalConnect(window, "destroy", Quit, NULL);

	gtk_selection_add_targets(window, GDK_SELECTION_PRIMARY,
			targetlist, sizeof(targetlist) / sizeof(GtkTargetEntry));

	GuiSignalConnect(window, "selection-clear-event",
					SelectionClear, &selection_urlstr);
	GuiSignalConnect(window, "selection-get",
					SelectionSend, &selection_urlstr);
	gtk_widget_realize(window);

	memset(notebook_list, '\0', sizeof(notebook_list));
	windata.toplist = &notebook_list[kTreeList];
	mainlist	= &notebook_list[kTreeList];
	GuiObjectSetUserData(window, &notebook_list[kTreeList]);

	guitl_create_icons(window);

	widget = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(window), widget);

	vbox = GTK_BOX(widget);

	widget = gtk_vbox_new(FALSE, 3);
	gtk_container_set_border_width(GTK_CONTAINER(widget), 2);
	gtk_box_pack_end(vbox, widget, FALSE, FALSE, 0);

	gui_build_msg_tmr(widget, &status_msg);
	label = gui_build_msg_tmr(widget, &error_msg);

	style = gtk_style_copy(gtk_widget_get_style(label));
	gdk_color_parse("Red", &style->fg[GTK_STATE_NORMAL]);
	gdk_color_parse("LightGray", &style->bg[GTK_STATE_NORMAL]);
	gtk_widget_set_style(label, style);

	error_msg.hdl_id = GuiSignalConnect(label, "expose-event",
						gui_error_msg_expose, NULL);
	GuiSignalHandlerBlock(label, error_msg.hdl_id);

	notebook = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), GTK_POS_TOP);
	gtk_box_pack_end(vbox, notebook, TRUE, TRUE, 2);

	gui_build_tree_list(notebook);
	gui_build_sorted_list(notebook);

	entry = gui_build_entry_select_browse();
	gtk_box_pack_end(vbox, entry, FALSE, FALSE, 0);

	menu_bar = gui_build_menu(window);
	gtk_box_pack_start(vbox, menu_bar, FALSE, FALSE, 0);

	GuiSignalConnect(notebook, "switch-page", notebook_page_switched, &windata);

	GuiTreeSignalConnect(&(notebook_list[kTreeList].store.tree),
		"button-press-event", gui_mouse_button_pressed, gui_popup_menu);

	GuiListSignalConnect(&(notebook_list[kSortedList].store.list),
		"button-press-event", gui_mouse_button_pressed, gui_popup_menu);

#if GTK_MAJOR_VERSION == 1

	GuiSignalConnect(window, "key-press-event",
			gui_sf10_show_popup_menu, gui_popup_menu);
#else
	GuiSignalConnect(window, "popup-menu",
			gui_sf10_show_popup_menu, gui_popup_menu);
#endif

	nscache_gui_started = TRUE;

	if (dbname)
	{
		gchar *str = utf8_from_str(dbname);

		if (str)
		{
			gtk_entry_set_text(entry_db, str);
			utf8_free(str);
		}

		g_free(dbname);
	}

	Open(entry_db);

	gtk_widget_show_all(window);

#ifndef CACHE_READONLY

	if (nscache_db.readonly)
	{
		gtk_widget_hide(gui_menu_widget[kMenuDelete]);
		gtk_widget_hide(gui_popup_widget[kMenuDelete]);
		gtk_widget_hide(gui_option[kOptionReadOnly].widget);
	}
#endif

} /* static void gui_init(char *dbname) */

int main(int argc, char **argv)
{
	int	error, i, visible;
	char	*dbname;

	INIT_LOCALE();

	error = cmd_setup(argc, argv, &dbname);

	if (error)
	{
		return error;
	}

#ifdef HAVE_LINK
	gprop_get_bool(kPropBoolSaveHardlink, &hardlink);
#endif
	gprop_get_bool(kPropBoolSaveSingleFolder, &single_folder);
	gprop_get_int(kPropIntSaveUrlToLocal, &url_to_local_selector);

#ifdef CONFIG_SWAP_COLUMNS

	column_tab_array[kColumnURL].position = kColumnURL;

	visible	= 0;
	i	= kColumnInfoLastIndex;

	do
	{
		ColumnTab *col_tab = &column_tab_array[i];

		if ( ! gprop_get_int(column_info[i].colnrprop, &col_tab->position))
		{
			col_tab->position = i;
		}

		visible += col_tab->position;
	}
	while (--i > kColumnURL);

	if (visible != 1+2+3+4+5+6+7+8)
	{
		ColumnTab *col_tab = &column_tab_array[kColumnInfoLastIndex];

		do
		{
			col_tab->position = 0;
		}
		while (--col_tab > &column_tab_array[kColumnURL]);
	}

#endif /* CONFIG_SWAP_COLUMNS */

	visible	= FALSE;
	i	= kColumnURL;

	do
	{
		ColumnTab *col_tab = &column_tab_array[i];

		col_tab->index = i;

#ifdef CONFIG_SWAP_COLUMNS
		if (col_tab->position == 0)
#endif
		{
			col_tab->position = i;
		}

#ifndef CONFIG_GTKCLIST_GTKCTREE
#ifdef CONFIG_COLUMN_POPUP
		col_tab->initcolnr = col_tab->position;
#endif
#endif
		column_tab[col_tab->position] = col_tab;

		if ( ! gprop_get_bool(column_info[i].show_column_prop, &col_tab->visible))
		{
			col_tab->visible = TRUE;
		}

		visible |= col_tab->visible;
	}
	while (++i <= kColumnInfoLastIndex);

	if (visible == FALSE)
	{
		column_tab_array[kColumnURL].visible = TRUE;
	}

	assign_default_viewers();

	gtk_init(&argc, &argv);
	gui_init(dbname);
	gtk_main();

	return 0;

} /* int main(int argc, char **argv) */

/* EOF */
