/***************************************************************************/
/* 		This code is part of Nscache - viewer of Netscape(tm)	   */
/*		browsers disk cache					   */
/*		Copyright (c) 1999,2000 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		modified 2005 by Harald Foerster			   */
/*		(harald_foerster@users.sourceforge.net)			   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#ifndef _re_h_
#define _re_h_

#ifdef HAVE_REGEX

#ifdef HAVE_REGEX_H

#ifdef HAVE_BSD_REGEX
#ifndef _REGEX_RE_COMP
#define _REGEX_RE_COMP
#endif
#endif

#include <sys/types.h>
#include <regex.h>

#endif /* HAVE_REGEX_H */


#if defined(HAVE_POSIX_REGEX)

typedef struct
{
	regex_t	preg;

} re_entry;

#elif defined(HAVE_V8_REGEX)

#ifndef NSUBEXP
#include <regexp.h>
#endif

#define REGEX_NO_IGNORE_CASE
#define REGEX_SEARCH_STR_LENGTH

typedef struct
{
	regexp	*preg;
	int	ignore_case;

} re_entry;

#elif defined(HAVE_BSD_REGEX)

#define REGEX_NO_IGNORE_CASE
#define REGEX_SEARCH_STR_LENGTH

#ifdef RE_SYNTAX_POSIX_EXTENDED
#define HAVE_BSD_REGEX_SYNTAX_OPTION	RE_SYNTAX_POSIX_EXTENDED
#endif

typedef struct
{
	char	*pattern;
	int	ignore_case;

} re_entry;

#elif defined(HAVE_GNU_REGEX)

#define REGEX_NO_IGNORE_CASE
#define REGEX_SEARCH_STR_LENGTH

#ifdef RE_SYNTAX_POSIX_EXTENDED
#define HAVE_GNU_REGEX_SYNTAX_OPTION	RE_SYNTAX_POSIX_EXTENDED
#endif

typedef struct
{
	struct re_pattern_buffer preg;
	struct re_registers	 regs;
	int			 ignore_case;

} re_entry;

#elif defined(HAVE_PCRE_REGEX)

#include <pcre.h>

#define REGEX_SEARCH_STR_LENGTH

typedef struct
{
	pcre		*preg;
	pcre_extra	*preg_extra;

} re_entry;

#endif /* #if defined(HAVE_POSIX_REGEX) .. #elif */

/* 'error' only valid if 're_make' returns NULL (never V8), must be freed */
extern re_entry	*re_make(const char *pattern, int ignore_case, char **error);
extern void	re_free(re_entry *ree);

#ifdef REGEX_SEARCH_STR_LENGTH

extern int re_pmatch(re_entry *ree, const char *str, int slen, int *end_offs);

#define RegexSearch(ree, str, slen, end_offs)	\
		re_pmatch(ree, str, slen, end_offs)

#else /* ! REGEX_SEARCH_NEED_STRING_LENGTH */

extern int re_pmatch(re_entry *ree, const char *str, int *end_offs);

#define RegexSearch(ree, str, slen, end_offs)	\
		re_pmatch(ree, str, end_offs)

#endif /* #ifdef REGEX_SEARCH_NEED_STRING_LENGTH .. #else */

#endif /* HAVE_REGEX */

#endif /* _re_h_ */
